############ Stress-testing PETSc ##########
"""
In this script we use PETSc to calculate the dispersion of the polaron.
"""
########### ################

# before we import numpy, set the number of threads for numpy multithreading
# using the OMP_NUM_THREADS environment variable
import numpy as np
from scipy.optimize import curve_fit

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os
import shutil
import time

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = (
    f"/mnt/c/Users/bobst/Documents/University_of_British_Columbia/"
    f"Physics/Mona_Berciu/Generalized_Green_function_cluster_expan"
    f"sion/new_lorentzfit"
)
sys.path.append(ggce_head_dir)
import ggce

# load the mpi communicator
from mpi4py import MPI

# load the Model and Executor classes to store parameters and manage the calculation
from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS
from ggce.executors.serial import SerialDenseExecutor
from ggce.utils.utils import lorentzian

M, N, M2, N2 = (
    int(sys.argv[1]),
    int(sys.argv[2]),
    int(sys.argv[3]),
    int(sys.argv[4]),
)

a, b, c = float(sys.argv[5]), float(sys.argv[6]), float(sys.argv[7])
w_init = float(sys.argv[8])

# run PETSc benchmark at cloud size cloud_ext with up to bosons_max bosons
cloud_ext_H = M
bosons_H = N

cloud_ext_tfd_H = M2
bosons_tfd_H = N2

coupling_H = a
Omega_H = b
temperature = c * Omega_H
hopping = 1.0
model_type1 = "Holstein"

# define parameters for the dispersion calculation
kgrid = np.linspace(0, np.pi / 20.0, 20)
eta = 0.05
next_k_offset_factor = 1.0
eta_step_div = 10.0

# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

results_dir = os.path.join(script_dir, f"pol_dis_HfT_{M}{N}_{M2}{N2}")

_CLEAN_START_ = True
if COMM.Get_rank() == 0:
    if not os.path.exists(results_dir):
        os.mkdir(results_dir)
    if os.path.exists(results_dir):
        if _CLEAN_START_:
            shutil.rmtree(results_dir)
            os.mkdir(results_dir)
        else:
            print(f"if you want a clean start, be careful!")

model = Model("finite_t_calc", "info", log_file=None)
model.set_parameters(
    hopping=hopping,
    temperature=temperature,
    lattice_constant=1.0,
    dimension=1,
    max_bosons_per_site=None,
)
model.add_coupling(
    model_type1,
    Omega=Omega_H,
    M=cloud_ext_H,
    N=bosons_H,
    M_tfd=cloud_ext_tfd_H,
    N_tfd=bosons_tfd_H,
    dimensionless_coupling=coupling_H,
)
model.finalize()

# initialize the Executor to manage the calculation
executor = ParallelSparseExecutorMUMPS(model, "info", mpi_comm=COMM)
# executor = SerialDenseExecutor(model, "info", mpi_comm=COMM)
executor.prime()

results = executor.dispersion(
    kgrid,
    w_init,
    eta,
    nmax=10000,
    eta_step_div=eta_step_div,
    next_k_offset_factor=next_k_offset_factor,
    peak_routine="change_w",
)
## this is a dict with keys k, w, A, ground_state (which is w of highest point) and weight (quasiparticle)
## so I should be fitting to pairs of (k, ground_state)
energy_k = [results[ii]["ground_state"] for ii in range(len(kgrid))]

# # only collect and output results on the head node
if COMM.Get_rank() == 0:

    # plot and save the Lorentz fit for diagnostic purposes
    for ii in range(len(kgrid)):
        fig, ax = plt.subplots(1, 1)
        loc = results[ii]["ground_state"]
        scale = results[ii]["weight"]
        ax.plot(
            results[ii]["w"],
            lorentzian(results[ii]["w"], loc, scale, eta),
            label="Lorentz fit",
        )
        ax.scatter(results[ii]["w"], results[ii]["A"], label="Calculation")
        ax.set_ylabel(r"$A(k,w)$", fontsize=20)
        ax.set_xlabel(r"$w$", fontsize=20)
        ax.tick_params(which="both", labelsize=16)
        ax.legend()
        plt.tight_layout()
        plt.savefig(
            os.path.join(results_dir, f"lorentzfit_k_{kgrid[ii]:.3f}.png"),
            format="png",
        )
        plt.close(fig)

    # process the solver output to extract time to solve and spectrum vals
    fig, ax = plt.subplots(1, 1)
    ax.scatter(kgrid, energy_k, label="PETSc")
    ax.set_ylabel(r"$E_P(k)$", fontsize=20)
    ax.set_xlabel(r"$k/\pi$", fontsize=20)
    ax.tick_params(which="both", labelsize=16)

    def quadratic_band(k, m_star, offset):

        return k**2 / (2.0 * m_star) - offset

    best_fit_pars, covmat = curve_fit(
        quadratic_band, kgrid, energy_k, p0=[0.5, energy_k[0]]
    )
    dense_kgrid = np.linspace(0, kgrid[-1], 100)
    ax.plot(
        dense_kgrid,
        quadratic_band(dense_kgrid, *best_fit_pars),
        label=f"$k^2$ fit",
    )

    # also output the energy data to disk for posterity / postprocess
    xx = np.array([kgrid, energy_k]).T
    np.savetxt(
        os.path.join(results_dir, f"polaron_dispersion.txt"),
        xx,
        header=f"momentum        E_P(k) with m_star / m_0 = {best_fit_pars[0] / 0.5041061511616247} (m_0 = 0.5041061511616247)",
        delimiter="    ",
    )

    plt.legend(bbox_to_anchor=(1, 1), loc="upper left", fontsize=16)
    plt.tight_layout()
    # plt.show()
    plt.savefig(
        os.path.join(results_dir, f"polaron_dispersion.png"),
        format="png",
        bbox_inches="tight",
    )

    ## also save all data just in case
    for ii, kval in enumerate(kgrid):
        xx = np.array([results[ii]["w"], results[ii]["A"]]).T
        np.savetxt(
            os.path.join(results_dir, f"all_results_k_{kval:.2f}.txt"),
            xx,
            header=f"omega            spectral function        for momentum = {kval}",
            delimiter="    ",
        )



############ Stress-testing PETSc ##########
"""
In this script we use PETSc reproduce some features of Fig. 2 in arxiv 2103.01972.
This is the final check that PETSc is reliable.
"""
########### ################

# before we import numpy, set the number of threads for numpy multithreading
# using the OMP_NUM_THREADS environment variable
import numpy as np

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os
import time

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(ggce_head_dir)
    import ggce

# load the mpi communicator
from mpi4py import MPI

# load the Model and Executor classes to store parameters and manage the calculation
from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS
from ggce.executors.serial import SerialDenseExecutor

# run convergence for cloud size up to cloud_max with up to bosons_max bosons
cloud_min = 3
cloud_max = 3
bosons_min = 30
bosons_max = 30

# fix parameter values same as ground truth from the https://export.arxiv.org/pdf/2103.01972
# article
Omega = np.array([10.0, 5.0, 2.0, 1.0, 0.5, 0.1, 0.05, 0.01])
coupling = np.linspace(0.0, 1.2, 30)
hopping = 1.0
eta = 0.005
next_lam_offset = 0.1
w0 = -2.1  # initial guess
model_type = "Holstein"

# define the k grid for the dispersion calculation
# since just looking for GS of Holstein, that is always at k = 0
kgrid = np.array([0])

# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

# start by creating a folder with the text and image output in run directory
# only head node make directories
results_dir = os.path.join(script_dir, "relb_gs_rlts")

if COMM.Get_rank() == 0:
    if not os.path.exists(results_dir):
        os.mkdir(results_dir)

# create a list to store gs energies
gs_energies = np.zeros(
    (
        cloud_max - cloud_min + 1,
        bosons_max - bosons_min + 1,
        len(Omega),
        len(coupling),
    )
)

# start the loop over all allowed M, N
fig, ax = plt.subplots(1, 1, figsize=(3, 2))
for m in range(cloud_min, cloud_max + 1):
    for n in range(bosons_min, bosons_max + 1):
        for ii, omega in enumerate(Omega):

            for jj, lam in enumerate(coupling):

                # initialize the model with pre-determined parameters
                model = Model()
                model.set_parameters(hopping=hopping)
                model.add_coupling(
                    model_type,
                    Omega=omega,
                    M=m,
                    N=n,
                    dimensionless_coupling=lam,
                )

                # initialize the Executor to manage the calculation
                executor = ParallelSparseExecutorMUMPS(
                    model, "info", mpi_comm=COMM
                )
                executor.prime()

                # re-define w0 so we are not wasting time
                if jj > 0:
                    w0 = (
                        gs_energies[m - cloud_min, n - bosons_min, ii, jj - 1]
                        - next_lam_offset
                    )

                # find the energy of the GS (polaron)
                results = executor.dispersion(kgrid, w0, eta, nmax=10000)
                gs_energy = results[0]["ground_state"]
                # add the gs energy to the matrix
                gs_energies[m - cloud_min, n - bosons_min, ii, jj] = gs_energy

            # # only collect and output results on the head node
            if COMM.Get_rank() == 0:
                # process the solver output to extract time to solve and spectrum vals
                ax.plot(
                    coupling,
                    gs_energies[m - cloud_min, n - bosons_min, ii, :],
                    label=rf"PETSc ($M={m}, N={n}, \Omega = {omega}$)",
                )
                ax.set_ylim(-3.1, -1.9)
                ax.set_xlim(0, 1.2)
                ax.set_ylabel(r"$E_{GS} / t$", fontsize=16)
                ax.set_xlabel(r"$\lambda$", fontsize=16)

                # also output the energy data to disk for posterity / postprocess
                xx = np.array(
                    [
                        coupling,
                        gs_energies[m - cloud_min, n - bosons_min, ii, :],
                    ]
                ).T
                np.savetxt(
                    os.path.join(
                        results_dir,
                        f"gs_energy_vs_lam_M_{m}_N_{n}_om_{omega}.txt",
                    ),
                    xx,
                    header=f"lambda        E_GS for Omega = {omega} and (M,N) = ({m},{n})",
                    delimiter="    ",
                )


if COMM.Get_rank() == 0:
    plt.legend(bbox_to_anchor=(1, 1), loc="upper left")
    # plt.show()
    plt.savefig(
        os.path.join(
            results_dir, f"gs_energy_vs_lam_M_{cloud_max}_N_{bosons_max}.png"
        ),
        format="png",
        bbox_inches="tight",
    )
    # also output the energy data to disk for posterity / postprocess
    xx = np.array(
        [
            coupling,
        ]
        + [
            gs_energies[m, n, l, :]
            for l in range(gs_energies.shape[2])
            for m in range(gs_energies.shape[0])
            for n in range(gs_energies.shape[1])
        ]
    ).T
    np.savetxt(
        os.path.join(
            results_dir,
            f"gs_energy_vs_lam_Mmax_{cloud_max}_Nmax_{bosons_max}.txt",
        ),
        xx,
        header=f"lambda        E_GS for various Omega and (M,N) arranged as Omega[0], M = 1, N = 1,..., M = 2, N = 1,... Omega[1]...",
        delimiter="    ",
    )



# TODO: entire module should be accelerated via C++

from copy import deepcopy
from collections import OrderedDict
from pathlib import Path

import json
import numpy as np
import pickle

from ggce.logger import logger
from ggce.model import Model
from ggce.engine.terms import Config, config_legal
from ggce.engine.equations import Equation, GreenEquation
from ggce.utils.utils import timeit
from ggce.utils.combinatorics import total_generalized_equations


# TODO: tagged for C++ acceleration
def config_space_gen(length, total_sum):
    """Generator for yielding all possible combinations of integers of
    length ``length`` that sum to ``total_sum``.

    .. warning::

        Not that cases such as ``length == 4`` and ``total_sum == 5`` such as
        ``[0, 0, 2, 3]`` still need to be screened out, since these do not
        correspond to valid f-functions.

    .. note::

        The algorithm to produce this code can be found `here <https://
        stackoverflow.com/questions/7748442/
        generate-all-possible-lists-of-length-n-that-sum-to-s-in-python>`_.

    Parameters
    ----------
    length : int
        The length of the array to produce.
    total_sum : int
        Constraints the sum over all of the elements of the array to equal this
        value.

    Yields
    ------
    tuple
    """

    if length == 1:
        yield (total_sum,)
    else:
        for value in range(total_sum + 1):
            for permutation in config_space_gen(length - 1, total_sum - value):
                r = (value,) + permutation
                yield r


def generate_all_legal_configurations(model):
    """Summary

    In one dimension, this is really easy. We can simply iterate over all
    0 < n <= N and 0 < m <= M. Things get much more complicated as the
    dimensionality increases. However, it is simplified somewhat since we
    always assume that M is a "square". We can still generate configurations
    that

    Parameters
    ----------
    model : TYPE
        Description
    """

    phonon_absolute_extent = model.phonon_absolute_extent
    max_phonons_per_site = model.phonon_max_per_site
    phonon_absolute_extent = model.phonon_absolute_extent
    phonon_extent = model.phonon_extent
    phonon_number = model.phonon_number
    n_phonon_types = model.n_phonon_types

    if model.hamiltonian.dimension > 1:
        logger.critical(">1 dimensions not yet implemented")

    nb_max = sum(phonon_number)
    config_dict = {n: [] for n in range(1, nb_max + 1)}
    for nb in range(1, nb_max + 1):
        for z in range(1, phonon_absolute_extent + 1):
            c = list(config_space_gen(z * n_phonon_types, nb))

            # Reshape everything properly:
            # This will only work for 1d!!!
            c = [np.array(cc).reshape(n_phonon_types, -1) for cc in c]

            # Deal with the multiple dimensions later!
            # TODO

            # Now, we get all of the LEGAL nvecs, which are those in
            # which at least a single phonon of any type is on both
            # edges of the cloud.
            tmp_legal_configs = [
                Config(cc)
                for cc in c
                if config_legal(
                    cc,
                    max_phonons_per_site=max_phonons_per_site,
                    phonon_extent=phonon_extent,
                    phonon_number=phonon_number,
                )
            ]

            # Extend the temporary config
            config_dict[nb].extend(tmp_legal_configs)

    return config_dict


class System:
    """Defines a list of Equations (a system of equations, so to speak) and
    operations on that system."""

    @property
    def model(self):
        return self._model

    @property
    def generalized_equations(self):
        """A list of equations constituting the system; in general form,
        meaning all except for the Green's function are not defined for
        specific delta values.

        Returns
        -------
        list
            A list of :class:`Equation` objects.
        """

        return self._generalized_equations

    @property
    def equations(self):
        return self._equations

    def _append_master_dictionary(self, eq):
        """Takes an equation and appends the master_f_arg_list dictionary."""

        if self._f_arg_list is None:
            self._f_arg_list = deepcopy(eq._f_arg_terms)
            return

        # Else, we append the terms
        for n_mat_id, l_deltas in eq._f_arg_terms.items():
            for delta in l_deltas:
                try:
                    self._f_arg_list[n_mat_id].append(delta)
                except KeyError:
                    self._f_arg_list[n_mat_id] = [delta]

    def _append_generalized_equation(self, n_phonons, config):

        eq = Equation.from_config(config, model=self._model)

        # Append a master dictionary at the System object level that
        # keeps track of all the f_arg values required for each value of
        # the config.
        self._append_master_dictionary(eq)

        # Finally, append the equation to a master list of generalized
        # equations/
        self._generalized_equations[n_phonons].append(eq)

    def _determine_unique_dictionary(self):
        """Sorts the master delta terms for easier readability and takes
        only unique delta terms."""

        for n_mat_id, l_deltas in self._f_arg_list.items():
            new_list = [
                np.array(xx)
                for xx in set([tuple(yy.tolist()) for yy in l_deltas])
            ]
            self._f_arg_list[n_mat_id] = new_list

    def _get_total_terms(self):
        """Predicts the total number of required specific equations needed
        to close the system."""

        return sum([len(ll) for ll in self._f_arg_list.values()])

    def _predict_total_terms(self):

        L = sum([len(val) for val in self._generalized_equations.values()])

        # Need to generalize this
        phonon_max_per_site = self._model.phonon_max_per_site
        n_phonon_types = self._model.n_phonon_types
        if n_phonon_types == 1 and phonon_max_per_site is None:
            # Plus one for the Green's function

            T = 1 + total_generalized_equations(
                self._model.phonon_extent,
                self._model.phonon_number,
                n_phonon_types,
            )

            if L == T:
                logger.info(
                    f"Predicted {L} generalized equations (agrees with "
                    "analytic formula)"
                )
            else:
                logger.error(
                    f"Predicted {T} generalized equations from analytic "
                    f"equation but {L} were generated. This will likely "
                    "in a critical error!"
                )
        else:
            logger.info(f"Predicted {L} generalized equations")

    def _initialize_generalized_equations(self, allowed_configs):

        self._generalized_equations = {n: [] for n in allowed_configs.keys()}

        # Generate all possible numbers of phonons consistent with n_max.
        for n_phonons, configs in allowed_configs.items():
            for config in configs:
                self._append_generalized_equation(n_phonons, config.config)

        eq = GreenEquation(model=self._model)
        self._append_master_dictionary(eq)

        # Only one Green's function, with "zero" phonons
        self._generalized_equations[0] = [eq]

        self._determine_unique_dictionary()
        self._predict_total_terms()

    def _initialize_equations(self):

        totals = self._get_total_terms()

        # Initialize the self._equations attribute's lists here since we know
        # all the keys:
        self._equations = {
            key: [] for key in self._generalized_equations.keys()
        }

        # Initialize the full set of equations
        for n_phonons, l_eqs in self._generalized_equations.items():
            for eq in l_eqs:
                n_mat_id = eq.index_term._get_phonon_config_id()
                l_deltas = self._f_arg_list[n_mat_id]
                for true_delta in l_deltas:
                    eq_copy = deepcopy(eq)
                    eq_copy._init_full(true_delta)
                    self._equations[n_phonons].append(eq_copy)

        L = sum([len(val) for val in self._equations.values()])

        if L == totals:
            logger.info(f"Generated {L} total equations")
        else:
            logger.error(
                f"Predicted {totals} equations from generalized form but {L} "
                f"were generated. This is likely a bug in the code that will "
                "result in a critical error!"
            )

    def _final_checks(self):
        """Runs a sanity check on the unique keys, which should equal the
        number of equations.
        """

        unique_short_identifiers = set()
        all_terms_rhs = set()
        for n_phonons, equations in self._equations.items():
            for eq in equations:
                unique_short_identifiers.add(eq.index_term.id())
                for term in eq._terms_list:
                    all_terms_rhs.add(term.id())

        if unique_short_identifiers == all_terms_rhs:
            logger.info("Closure checked and valid")
        else:
            # logger.error("Invalid closure!")
            # logger.error(unique_short_identifiers - all_terms_rhs)
            # logger.error(all_terms_rhs - unique_short_identifiers)
            logger.critical("Critical error due to invalid closure.")

    def _save_model(self):
        """Saves the model attribute to disk. Will not overwrite an existing
        file."""

        if self._root is None:
            return

        path = Path(self._root) / Path("model.json")
        m = json.loads(self._model.to_json())
        with open(path, "w") as f:
            json.dump(m, f, indent=4, sort_keys=False)

    def checkpoint(self):
        """Runs the checkpoint protocol to attempt and save the current
        state of the system. This method will only run if a root directory is
        defined at class instantiation. Note that existing checkpoints will
        never be overwritten.

        If the number of equations is larger than a pre-specified cutoff
        (by default 1,000,000), then checkpointing raises a warning and
        does not work.
        """

        if self._root is None:
            logger.warning(
                "root not provided to System - System checkpointing disabled"
            )
            return

        # check how many equations there are -- if too many, checkpointing
        # will be disabled
        try:
            L = sum([len(val) for val in self._equations.values()])
            if L > self._chkpt_lim:
                logger.warning(
                    f"\nMatrices have size {L}, which is larger than the"
                    f" checkpoint limit {self._chkpt_lim:.0f}.\nPickle "
                    f"performance expected to be slow for such large"
                    f" files. System checkpointing is disabled.\n"
                    f"To override the preset, pass a value for chkpt_lim"
                    f" to System object."
                )
                return
        except AttributeError:  # just means no equations generated yet
            pass

        for attr in ["generalized_equations", "f_arg_list", "equations"]:
            obj = eval(f"self._{attr}")
            path = Path(self._root) / Path(f"{attr}.pkl")
            if obj is not None and not path.exists():
                pickle.dump(
                    obj,
                    open(path, "wb"),
                    protocol=pickle.HIGHEST_PROTOCOL,
                )
                logger.info(f"Checkpoint saved: {attr}")

    @classmethod
    def from_checkpoint(cls, root):
        """Reloads the state of the System from the saved pickle file
        checkpoint.

        Parameters
        ----------
        root : os.PathLike
            Checkpoint directory
        """

        path = Path(root) / Path("model.json")
        if not path.exists():
            logger.critical(f"Model checkpoint file {path} does not exist")

        with open(path, "r") as f:
            model = Model.from_dict(json.load(f))

        generalized_equations = None
        path = Path(root) / Path("generalized_equations.pkl")
        if path.exists():
            generalized_equations = pickle.load(open(path, "rb"))

        f_arg_list = None
        path = Path(root) / Path("f_arg_list.pkl")
        if path.exists():
            f_arg_list = pickle.load(open(path, "rb"))

        equations = None
        path = Path(root) / Path("equations.pkl")
        if path.exists():
            equations = pickle.load(open(path, "rb"))

        return cls(
            model=model,
            generalized_equations=generalized_equations,
            f_arg_list=f_arg_list,
            equations=equations,
            root=str(root),
        )

    def __init__(
        self,
        model=None,
        generalized_equations=None,
        f_arg_list=None,  # TODO: pool f_arg_list into generalized_equations
        equations=None,
        root=None,
        mpi_comm=None,
        chkpt_lim=1e6,
        autoprime=True,
    ):
        self._root = root
        if self._root is not None:
            Path(root).mkdir(exist_ok=True, parents=True)

        self._model = deepcopy(model)
        self._save_model()

        self._generalized_equations = generalized_equations
        self._f_arg_list = f_arg_list
        self._equations = equations
        self._chkpt_lim = chkpt_lim

        if autoprime:
            # Get all of the allowed configurations - additionally checkpoint
            # the generalized configurations in a root directory if its
            # provided
            self._prime_system()
        else:
            logger.warning(
                "System has not been primed. Load from disk "
                "or prime explicitly with _prime_system."
            )

    def _prime_system(self):
        """Helper function that facilitates the construction of configs,
        generalized and total equations. By default it is called during
        init, but it can be disabled by passing autoprime=False, in case a
        user wants to initialize the system from disk or save initialization
        for later."""
        if self._generalized_equations is None:
            with timeit(logger.debug, "Legal configurations generated"):
                confs = generate_all_legal_configurations(self._model)
            with timeit(logger.debug, "Generalized equations initialized"):
                self._initialize_generalized_equations(confs)
        else:
            logger.info("Generalized equations reloaded from state")
        self.checkpoint()

        if self._equations is None:
            with timeit(logger.debug, "Equations initialized"):
                self._initialize_equations()
        else:
            logger.info("Equations reloaded from state")
        self.checkpoint()

        with timeit(logger.debug, "Final checks"):
            self._final_checks()

    def visualize(self, generalized=True, full=True, coef=None):
        """Allows for easy visualization of the closure. Note this isn't
        recommended when there are greater than 10 or so equations, since
        it will be very difficult to see everything.

        Parameters
        ----------
        generalized : bool
            If True, prints information on the generalized equations, else
            prints the full equations. Default is True.
        full : bool
            If True, prints information about the argument of g(...) and the
            exponential shift in addition to the n-vector and f-argument.
            Else, just prints the latter two. Default is True.
        coef : list, optional
            If not None, actually evaluates the terms at the value of the
            (k, w) coefficient. Default is None.
        """

        eqs_dict = (
            self._generalized_equations if generalized else self._equations
        )
        od = OrderedDict(sorted(eqs_dict.items(), reverse=True))
        for n_phonons, eqs in od.items():
            print(f"{n_phonons}")
            print("-" * 60)
            for eq in eqs:
                eq.visualize(full=full, coef=coef)
            print("\n")

    def get_basis(self, full_basis=False):
        """Prepares the solver-specific information.

        Returns the non-zero elements of the matrix in the following format.
        The returned quantity is a dictionary indexed by the order of the
        hierarchy (in this case, the number of phonons contained). Each
        element of this dictionary is another dictionary, with the keys being
        the index term identifier (basically indexing the row of the matrix),
        and the values a list of tuples, where the first element of each
        is the identifier (a string) and the second is a callable function of
        k and omega, representing the coefficient at that point.

        Parameters
        ----------
        full_basis : bool, optional
            If True, returns the full basis mapping. If False, returns the
            local basis mapping, which is used in the continued fraction
            solver. (The default is False).

        Returns
        -------
        dict
            The dictionary objects containing the basis.
        """

        # The basis object maps each unique identifier to a unique number.
        # The local_basis object maps each unique identifier to a unique
        # number within the manifold of some number of phonons.
        basis = dict()

        if full_basis:

            # Set the overall basis. Each unique identifier gets its own .
            cc = 0
            for _, equations in self._equations.items():
                for eq in equations:
                    basis[eq.index_term.id()] = cc
                    cc += 1

        else:

            # Set the local basis, in which each identifier gets its own
            # relative to the n-phonon manifold.
            for n_phonons, list_of_equations in self._equations.items():
                basis[n_phonons] = {
                    eq.index_term.id(): ii
                    for ii, eq in enumerate(list_of_equations)
                }

        return basis



from .logger import logger  # noqa
from .model import Model  # noqa
from .engine.system import System  # noqa
from .executors.solvers import SparseSolver, DenseSolver  # noqa


# DO NOT CHANGE ANYTHING BELOW THIS
__version__ = '0.1'
# DO NOT CHANGE ANYTHING ABOVE THIS



# Run with mpiexec -np 4 python3 petsc4py_hello_world.py

from petsc4py import PETSc

rank = PETSc.COMM_WORLD.Get_rank()
size = PETSc.COMM_WORLD.Get_size()

if rank == 0:
    print("Note: PETSc scalar type should be complex!")
    print("If not, you've installed petsc incorrectly.")
    print(PETSc.ScalarType)

print(f"On rank {rank}/{size}")

# Should see something like
# On rank 1/4
# On rank 3/4
# On rank 0/4
# On rank 2/4

A = PETSc.Mat()
A.create(PETSc.COMM_WORLD)
A.setSizes([100, 100])



from numpy import prod
from scipy.special import comb


def generalized_equations_combinatorics_term(m, n):
    """The total number of generalized equations is given by the exact
    relation

    sum_{m, n = 1}^{M, N} c_{m, n}

    where c_{m, n} is given by this equation, and is equal to

    * 1 if m = 1 or n = 2
    * (m + n - 3) choose (n - 2) otherwise
    """

    if m == 1 or n == 2:
        return 1

    return comb(m + n - 3, n - 2, exact=True)


def total_generalized_equations(M, N, nbt):
    """Gets the total number of generalized equations as predicted by the
    combinatorics equation described in
    generalized_equations_combinatorics_term.
    """

    bosons = [
        sum(
            [
                sum(
                    [
                        generalized_equations_combinatorics_term(m, n)
                        for n in range(1, N[bt] + 1)
                    ]
                )
                for m in range(1, M[bt] + 1)
            ]
        )
        for bt in range(nbt)
    ]
    return int(prod(bosons))



from abc import ABC, abstractmethod
from pathlib import Path
import pickle
import warnings

import numpy as np
from scipy import linalg
from scipy.sparse import coo_matrix
from scipy.sparse.linalg import spsolve
from tqdm import tqdm

from ggce.logger import logger, disable_logger
from ggce.utils.utils import float_to_list, chunk_jobs
from ggce.engine.system import System
from ggce.utils.physics import G0_k_omega


BYTES_TO_MB = 1048576


class Solver(ABC):
    @property
    def system(self):
        return self._system

    @property
    def root(self):
        return self._root

    @property
    def basis(self):
        return self._basis

    @property
    def mpi_comm(self):
        return self._mpi_comm

    @property
    def mpi_rank(self):
        if self._mpi_comm is not None:
            return self._mpi_comm.Get_rank()
        return 0

    @property
    def mpi_world_size(self):
        if self._mpi_comm is not None:
            return self._mpi_comm.Get_size()
        return 1

    def __init__(self, system=None, root=None, basis=None, mpi_comm=None):
        self._system = system
        self._root = root
        self._mpi_comm = mpi_comm
        self._basis = basis

        if self._system is None and self._root is None:
            logger.critical("Either system, root or both must be provided")

        if self._root is not None:
            # We allow checkpointing
            self._root = Path(self._root)
            self._results_directory = self._root / Path("results")
            self._results_directory.mkdir(exist_ok=True, parents=True)

        else:
            logger.warning(
                "root not provided to Solver - Solver checkpointing disabled"
            )
            self._results_directory = None

        if self._system is None:
            # Attempt to load the system from its checkpoint... the system
            # will now be initialized or an error will be thrown
            self._system = System.from_checkpoint(self._root)

        # Force checkpoint the system, which at this point must be initialized
        with disable_logger():
            self._system.checkpoint()

        if self._root is not None:
            logger.info(f"System checkpointed to '/{self._root}'")

    def get_jobs_on_this_rank(self, jobs):
        """Gets the jobs assigned to this rank. Note this method silently
        behaves as it should when the world size is 1 (or there's no MPI
        communicator).

        Parameters
        ----------
        jobs : list
            The jobs to chunk.

        Returns
        -------
        list
            The jobs assigned to this rank.
        """

        if self.mpi_comm is None:
            return jobs

        return chunk_jobs(jobs, self.mpi_world_size, self.mpi_rank)

    @staticmethod
    def _k_omega_eta_to_str(k, omega, eta):
        # Note this will have to be redone when k is a vector in 2 and 3D!
        return f"{k:.10f}_{omega:.10f}_{eta:.10f}"

    @abstractmethod
    def _pre_solve(self):
        ...

    @abstractmethod
    def _post_solve(self):
        ...

    @abstractmethod
    def solve(self, k, w, eta):
        """Takes, ``k, w, eta`` and returns the Green's function."""
        ...

    @abstractmethod
    def greens_function(self, k, w, eta, pbar=False):
        ...


class BasicSolver(Solver):
    def _pre_solve(self, k, w, eta):
        result = None
        path = None
        if self._results_directory is not None:
            ckpt_path = f"{self._k_omega_eta_to_str(k, w, eta)}.pkl"
            path = self._results_directory / Path(ckpt_path)
            if path.exists():
                result = np.array(pickle.load(open(path, "rb")))
        return result, path

    def _post_solve(self, G, k, w, path):
        if -G.imag / np.pi < 0.0:
            logger.error(f"A(k,w) < 0 at k, w = ({k:.02f}, {w:.02f}")
        if self._results_directory is not None:
            pickle.dump(G, open(path, "wb"), protocol=pickle.HIGHEST_PROTOCOL)

    def greens_function(self, k, w, eta, pbar=False):
        """Solves for the greens_function in serial or in parallel, depending
        on whether MPI_COMM is provided to the Solver at instantiation.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
        numpy.ndarray
            The resulting greens_function of shape ``nk`` by ``nw``.
        """

        k = float_to_list(k)
        w = float_to_list(w)

        # This orders the jobs such that when constructed into an array, the
        # k-points are the rows and the w-points are the columns after reshape
        jobs = [(_k, _w) for _k in k for _w in w]
        jobs_on_rank = self.get_jobs_on_this_rank(jobs)

        s = []
        for (_k, _w) in tqdm(jobs_on_rank, disable=not pbar):
            s.append(self.solve(_k, _w, eta))

        if self.mpi_comm is not None:
            all_results = self.mpi_comm.gather(s, root=0)

            # Only rank 0 returns the result
            if self.mpi_rank == 0:
                # all_results is a list of arrays of different length
                # need to parse it properly into an array
                all_results = [
                    xx[ii] for xx in all_results for ii in range(len(xx))
                ]
                s = np.array([xx for xx in all_results])
                return s.reshape(len(k), len(w))
            else:
                return None

        return np.array(s).reshape(len(k), len(w))


class SparseSolver(BasicSolver):
    """A sparse, serial solver for the Green's function. Useful for when the
    calculation being performed is quite cheap. Note that there are a variety
    of checkpointing features automatically executed when paths are provided.

    - When a ``System`` and path are provided, the system's root directory
      is overwritten and that object immediately checkpointed to the provided
      directory.

    - If only a ``System`` object is provided, no checkpointing will be
      performed.

    - If only a path is provided, the solver will attempt to instantiate the
      ``System`` object. Checkpointing will the proceed as normal.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._basis is None:
            self._basis = self._system.get_basis(full_basis=True)

    def _sparse_matrix_from_equations(self, k, w, eta):
        """This function iterates through the GGCE equations dicts to extract
        the row, column coordiante and value of the nonzero entries in the
        matrix. This is subsequently used to construct the parallel sparse
        system matrix. This is exactly the same as in the Serial class: however
        that method returns X, v whereas here we need row_ind/col_ind_dat.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
        list, list, list
            The row and column coordinate lists, as well as a list of values of
            the matrix that are nonzero.
        """

        row_ind = []
        col_ind = []
        dat = []

        total_bosons = np.sum(self._system.model.phonon_number)
        for n_bosons in range(total_bosons + 1):
            for eq in self._system.equations[n_bosons]:
                row_dict = dict()
                index_term_id = eq.index_term.id()
                ii_basis = self._basis[index_term_id]

                for term in eq._terms_list + [eq.index_term]:
                    jj = self._basis[term.id()]
                    try:
                        row_dict[jj] += term.coefficient(k, w, eta)
                    except KeyError:
                        row_dict[jj] = term.coefficient(k, w, eta)

                row_ind.extend([ii_basis for _ in range(len(row_dict))])
                col_ind.extend([key for key, _ in row_dict.items()])
                dat.extend([value for _, value in row_dict.items()])

        # estimate sparse matrix memory usage
        # (complex (16 bytes) + int (4 bytes) + int) * nonzero entries
        est_mem_used = 24 * len(dat) / BYTES_TO_MB
        logger.debug(f"Estimated memory needed is {est_mem_used:.02f} MB")

        return row_ind, col_ind, dat

    def _scaffold(self, k, w, eta):
        """Prepare the X, v sparse representation of the matrix to solve.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float, optional
            The artificial broadening parameter of the calculation.

        Returns
        -------
        csr_matrix, csr_matrix
            Sparse representation of the matrix equation to solve, X and v.
        """

        row_ind, col_ind, dat = self._sparse_matrix_from_equations(k, w, eta)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=DeprecationWarning)
            proto_matr = (
                np.array(dat, dtype=np.complex64),
                (np.array(row_ind), np.array(col_ind)),
            )

        X = coo_matrix(proto_matr).tocsr()

        size = (X.data.size + X.indptr.size + X.indices.size) / BYTES_TO_MB

        logger.debug(f"Memory usage of sparse X is {size:.01f} MB")

        # Initialize the corresponding sparse vector
        # {G}(0)
        row_ind = np.array([self._basis["{G}[0.0]"]])
        col_ind = np.array([0])
        v = coo_matrix(
            (
                np.array(
                    [self._system.equations[0][0].bias(k, w, eta)],
                    dtype=np.complex64,
                ),
                (row_ind, col_ind),
            )
        ).tocsr()

        return X, v

    def solve(self, k, w, eta):
        """Solve the sparse-represented system for some given point
        :math:`(k, \\omega, \\eta)`. Note that if the spectral function is
        negative results will not be checkpointed.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
        numpy.ndarray
            The value of :math:`G(k, \\omega; \\eta)`.
        """

        result, path = self._pre_solve(k, w, eta)
        if result is not None:
            return result

        # Solution ------------------------------------------------------------
        X, v = self._scaffold(k, w, eta)
        res = spsolve(X, v)
        G = res[self._basis["{G}[0.0]"]]
        # Solution Done -------------------------------------------------------

        self._post_solve(G, k, w, path)

        return np.array(G)


class DenseSolver(BasicSolver):
    """A sparse, dense solver for the Green's function. Note that there are a
    variety of checkpointing features automatically executed when paths are
    provided. Uses the SciPy dense solver engine to solve for G(k, w) in
    serial. This method uses the continued fraction approach,

    .. math:: R_{n-1} = (1 - \\beta_{n-1}R_{n})^{-1} \\alpha_{n-1}

    with

    .. math:: R_n = \\alpha_n

    and

    .. math:: f_n = R_n f_{n-1}
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._basis is None:
            self._basis = self._system.get_basis(full_basis=False)

    def _fill_matrix(self, k, w, n_phonons, shift, eta):

        n_phonons_shift = n_phonons + shift

        equations_n = self._system.equations[n_phonons]

        # Initialize a matrix to fill
        d1 = len(self._basis[n_phonons])
        d2 = len(self._basis[n_phonons + shift])
        A = np.zeros((d1, d2), dtype=np.complex64)

        # Fill the matrix of coefficients
        for ii, eq in enumerate(equations_n):
            index_term_id = eq.index_term.id()
            ii_basis = self._basis[n_phonons][index_term_id]
            for term in eq._terms_list:
                if term.config.total_phonons != n_phonons_shift:
                    continue
                t_id = term.id()
                jj_basis = self._basis[n_phonons_shift][t_id]
                A[ii_basis, jj_basis] += term.coefficient(k, w, eta)

        return A

    def _get_alpha(self, k, w, n_phonons, eta):
        return self._fill_matrix(k, w, n_phonons, -1, eta)

    def _get_beta(self, k, w, n_phonons, eta):
        return self._fill_matrix(k, w, n_phonons, 1, eta)

    def solve(self, k, w, eta):
        """Solve the dense-represented system.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.

        Returns
        -------
        np.ndarray
            The value of G.
        """

        result, path = self._pre_solve(k, w, eta)
        if result is not None:
            return result

        # Solution ------------------------------------------------------------
        total_phonons = np.sum(self._system.model.phonon_number)
        for n_phonons in range(total_phonons, 0, -1):

            # Special case of the recursion where R_N = alpha_N.
            if n_phonons == total_phonons:
                R = self._get_alpha(k, w, n_phonons, eta)
                continue

            # Get the next loop's alpha and beta values
            beta = self._get_beta(k, w, n_phonons, eta)
            alpha = self._get_alpha(k, w, n_phonons, eta)

            # Compute the next R
            X = np.eye(beta.shape[0], R.shape[1]) - beta @ R
            R = linalg.solve(X, alpha)

        a = self._system.model.lattice_constant
        t = self._system.model.hopping
        G0 = G0_k_omega(k, w, a, eta, t)

        beta0 = self._get_beta(k, w, 0, eta)
        G = (G0 / (1.0 - beta0 @ R)).squeeze()
        G = np.array(G, dtype=np.complex64)
        # Solution Done -------------------------------------------------------

        self._post_solve(G, k, w, path)

        return G



import pytest

import numpy as np

from ggce.utils import physics


def g0_delta_omega(delta, omega, a, eta, tf, sgn=1.0, e0=0.0):
    """This is the empirically correct g0_delta_omega. Any changes to the
    function (for e.g. speed improvements) defined in physics should return
    the same result as this function."""

    if tf == 0.0:
        if delta != 0:
            return 0.0
        return 1.0 / (omega + eta * 1j)

    omega = omega + eta * 1j
    B = 2.0 * np.abs(tf)
    x = (omega - e0) / B
    prefactor = -sgn * 1j / np.sqrt(B**2 - (omega - e0) ** 2)
    t1 = (-x + sgn * 1j * np.sqrt(1.0 - x**2)) ** np.abs(delta * a)
    return t1 * prefactor


@pytest.mark.parametrize("d", [-1, 100])
@pytest.mark.parametrize("o", [-0.5, -50.0, 50.0])
@pytest.mark.parametrize("e", [0.1, 1e-8])
@pytest.mark.parametrize("t", [0.0, 50.0])
def test_g0_delta_omega(d, o, e, t):
    assert g0_delta_omega(d, o, 1.0, e, t) == physics.g0_delta_omega(
        d, o, 1.0, e, t
    )


def G0_k_omega(k, omega, a, eta, tf):
    """This is the empirically correct G0_k_omega. Any changes to the
    function (for e.g. speed improvements) defined in physics should return
    the same result as this function."""

    epsilon_k = -2.0 * tf * np.cos(k * a)
    return 1.0 / (omega + 1j * eta - epsilon_k)


@pytest.mark.parametrize("k", [0.0, 0.5 * np.pi, np.pi])
@pytest.mark.parametrize("o", [-0.5, -50.0, 50.0])
@pytest.mark.parametrize("e", [0.1, 1e-8])
@pytest.mark.parametrize("t", [0.0, 50.0])
def test_G0_k_omega(k, o, e, t):
    assert G0_k_omega(k, o, 1.0, e, t) == physics.G0_k_omega(k, o, 1.0, e, t)



# before we import numpy, set the number of threads for numpy multithreading
# typically most gains are achieved by 4-6 threads
import numpy as np

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(ggce_head_dir)
    import ggce

from mpi4py import MPI

from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS

# run PETSc benchmark at cloud size cloud_ext with up to bosons_max bosons
cloud_ext = 8
bosons_min = 10
bosons_max = 10
bosons_step = 1
bosons_array = np.arange(bosons_min, bosons_max + 1, 1)

# set system parameters
Omega = 1.25
coupling = 2.5
hopping = 0.1
eta = 0.005

# pick a few points to do a calculation for parallel optimization
pts = 5
wgrid = np.linspace(-5.5, -2.5, pts)
k = 0.5 * np.pi

# loop over the range of allowed N_bosons to execute one computation (k,w)
# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

parallel_times = np.zeros((len(bosons_array), pts))
parallel_greens = np.zeros((len(bosons_array), pts), dtype="complex")

for ii, n in enumerate(bosons_array):

    model = Model()
    model.set_parameters(hopping=hopping)
    model.add_coupling(
        "EdwardsFermionBoson",
        Omega=Omega,
        M=cloud_ext,
        N=n,
        dimensionless_coupling=coupling,
    )

    # repeat the calculation iters times for error bar generation
    for jj, w in enumerate(wgrid):
        executor = ParallelSparseExecutorMUMPS(model, "info", mpi_comm=COMM)
        executor.prime()

        result = executor.solve(k, w, eta)

        if COMM.Get_rank() == 0:
            G, meta = result
            parallel_times[ii, jj] = meta["time"][-1]
            parallel_greens[ii, jj] = G

## now we print to disk
if COMM.Get_rank() == 0:

    # first calculate averages and standard error
    time_average_parallel = np.average(parallel_times, axis=-1)
    std_error_parallel = [
        np.std(parallel_times[ii, :], ddof=1)
        / np.sqrt(len(parallel_times[ii, :]))
        for ii in range(len(bosons_array))
    ]

    xx = np.array([time_average_parallel, std_error_parallel]).T
    np.savetxt(
        os.path.join(
            script_dir,
            f"speed_benchmark_M_{str(cloud_ext)}_N_{bosons_min}_{bosons_max}.txt",
        ),
        xx,
        header=f"parallel time (s)    parallel time error (s)",
        fmt="%.3f",
        delimiter="    ",
    )



# before we import numpy, set the number of threads for numpy multithreading
# typically most gains are achieved by 4-6 threads
import numpy as np

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os
import shutil

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(ggce_head_dir)
    import ggce

from mpi4py import MPI

from ggce.model import Model  # noqa: E402
from ggce.executors.serial import SerialDenseExecutor
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS

_CLEAN_START_ = True

# run PETSc benchmark at cloud size cloud_ext with up to bosons_max bosons
cloud_ext_H = 3
bosons_H = 9

cloud_ext_tfd_H = 3
bosons_tfd_H = 5

Omega_H = 1.0
coupling_H = 1.0
temperature = 0.4 * Omega_H
hopping = 1.0
eta = 0.005
model_type1 = "Holstein"

# define the k grid for the dispersion calculation
kgrid = np.array([0])
# pick several points to generate error bars
wgrid = np.linspace(-6.0, 6.0, 10)

# loop over the range of allowed N_bosons to execute one computation (k,w)
# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

results_dir = os.path.join(script_dir, "repr_fig7_dmrg")

if COMM.Get_rank() == 0:
    if not os.path.exists(results_dir):
        os.mkdir(results_dir)
    if os.path.exists(results_dir):
        if _CLEAN_START_:
            shutil.rmtree(results_dir)
            os.mkdir(results_dir)
        else:
            print(f"if you want a clean start, be careful!")

model = Model("repr_fig7_dmrg", "info", log_file=None)
model.set_parameters(
    hopping=hopping,
    temperature=temperature,
    lattice_constant=1.0,
    dimension=1,
    max_bosons_per_site=None,
)
model.add_coupling(
    model_type1,
    Omega=Omega_H,
    M=cloud_ext_H,
    N=bosons_H,
    M_tfd=cloud_ext_tfd_H,
    N_tfd=bosons_tfd_H,
    dimensionless_coupling=coupling_H,
)
model.finalize()

# print model parameters for the log
if COMM.Get_rank() == 0:
    model.visualize()

executor = ParallelSparseExecutorMUMPS(model, "info", mpi_comm=COMM)
executor.prime()

for ii, k in enumerate(kgrid):
    for jj, w in enumerate(wgrid):

        g_func, exitcodes_dict = executor.solve(k, w, eta)
        a_func = -g_func.imag / np.pi
        ## now we print to disk (and plot, if have matplotlib) the two timings to compare
        if COMM.Get_rank() == 0:

            times = exitcodes_dict["time"][0]
            mumps_exit_codes = exitcodes_dict["mumps_exit_code"][0]
            mumps_mem_tot = exitcodes_dict["mumps_mem_tot"][0]
            tol_excess = exitcodes_dict["manual_tolerance_excess"][0]

            # also output the spectrum data to disk for posterity / postprocess
            # create first entries which are momentum
            xx = np.array(
                [
                    [
                        k,
                        w,
                        a_func,
                        times,
                        mumps_exit_codes,
                        mumps_mem_tot,
                        tol_excess,
                    ]
                ]
            )
            # create the output file
            # if this is the first time then create the file with the heading
            result_file = os.path.join(results_dir, f"repr_fig7_dmrg.out")
            if not os.path.exists(result_file):
                np.savetxt(
                    result_file,
                    xx,
                    header=f"momentum                    omega                         "
                    f"spectral func               time_to_compute (sec)    "
                    f"MUMPS convergence code (0 = good)  MUMPS total memory used (Gb)  "
                    f"Tolerance excess  (|r| - rtol*|b|)",
                    delimiter="    ",
                )
            else:
                with open(result_file, "a") as datafile:
                    # datafile.write("\n")
                    np.savetxt(datafile, xx, delimiter="    ")

## plot the spectrum function cut for fig. 7 comparison
if COMM.Get_rank() == 0:
    result_file = os.path.join(results_dir, f"repr_fig7_dmrg.out")
    ## now let's visualize the spectrum and see what we have
    (
        k_vals,
        w_vals,
        a_func,
        times,
        mumps_exit_codes,
        mumps_mem_tot,
        tol_excess,
    ) = np.loadtxt(result_file, unpack=True, skiprows=1)
    # here it will be only a single k-slice
    fig, ax = plt.subplots(1, 1, figsize=(6, 6))
    ax.plot(
        w_vals, a_func, "b", label=rf"GGCE ($M={cloud_ext_H}, N={bosons_H}$)"
    )
    ax.set_ylabel(f"$A(0, \omega)$", size=16)
    ax.set_xlabel("$\omega / \omega_0$", size=16)
    ax.tick_params(labelsize=14)
    ax.set_xticks([-3, 0, 3])
    ax.set_yticks([0, 1, 2])
    ax.set_xlim(-6, 6)
    ax.set_ylim(-0.05, 2.5)
    plt.legend(fontsize=14)
    plt.tight_layout()
    plt.savefig(
        os.path.join(
            results_dir, f"repr_fig7_dmrg_M_{cloud_ext_H}_N_{bosons_H}.png"
        ),
        format="png",
    )



import json
import os
import pytest
import shutil

import numpy as np

from pathlib import Path
import pickle

from ggce import Model, System, SparseSolver

model_h = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.0
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=4,
        phonon_number=7,
        dimensionless_coupling_strength=0.6,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_p = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.0
    ),
    "model_add_params": dict(
        coupling_type="Peierls",
        phonon_frequency=2.35,
        phonon_extent=3,
        phonon_number=4,
        dimensionless_coupling_strength=1.3,
    ),
    "k": 1.46,
    "eta": 0.5,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_hp = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.0
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=4,
        dimensionless_coupling_strength=0.6,
    ),
    "model_add_params2": dict(
        coupling_type="Peierls",
        phonon_frequency=0.69,
        phonon_extent=2,
        phonon_number=3,
        dimensionless_coupling_strength=2.3,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_ht = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.6
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=3,
        phonon_extent_tfd=2,
        phonon_number_tfd=4,
        dimensionless_coupling_strength=0.6,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_pt = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=1.3
    ),
    "model_add_params": dict(
        coupling_type="Peierls",
        phonon_frequency=2.35,
        phonon_extent=3,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=4,
        dimensionless_coupling_strength=1.3,
    ),
    "k": 1.46,
    "eta": 0.5,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_hpt = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=2.1
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=2,
        dimensionless_coupling_strength=0.6,
    ),
    "model_add_params2": dict(
        coupling_type="Peierls",
        phonon_frequency=0.69,
        phonon_extent=2,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=2,
        dimensionless_coupling_strength=2.3,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}


@pytest.mark.parametrize(
    "p",
    [
        model_h,
        model_p,
        model_hp,
        model_ht,
        model_pt,
        model_hpt,
    ],
)
def test_sys_chkpt(p):
    model = Model.from_parameters(**p["model_params"])
    model.add_(**p["model_add_params"])
    # accommodate double-parameter case
    try:
        model.add_(**p["model_add_params2"])
    except KeyError:
        pass

    root = p["root_sys"]
    sys = System(model=model, root=root)
    sys.checkpoint()

    sys_disk = System.from_checkpoint(root)

    assert sys._f_arg_list == sys_disk._f_arg_list

    # iterate through the tree of equations to extract unique identifiers
    # compare identifiers between the two Systems
    unique_short_identifiers = set()
    all_terms_rhs = set()
    for n_phonons, equations in sys._equations.items():
        for eq in equations:
            unique_short_identifiers.add(eq.index_term.id())
            for term in eq._terms_list:
                all_terms_rhs.add(term.id())

    unique_short_identifiers_disk = set()
    all_terms_rhs_disk = set()
    for n_phonons, equations in sys_disk._equations.items():
        for eq in equations:
            unique_short_identifiers_disk.add(eq.index_term.id())
            for term in eq._terms_list:
                all_terms_rhs_disk.add(term.id())

    assert unique_short_identifiers == all_terms_rhs_disk

    shutil.rmtree(root)


@pytest.mark.parametrize(
    "p",
    [
        model_h,
        model_p,
        model_hp,
        model_ht,
        model_pt,
        model_hpt,
    ],
)
def test_model_chkpt(p):
    model = Model.from_parameters(**p["model_params"])
    model.add_(**p["model_add_params"])
    # accommodate double-parameter case
    try:
        model.add_(**p["model_add_params2"])
    except KeyError:
        pass
    root = p["root_sys"]
    System(model=model, root=root, autoprime=False)

    with open(os.path.join(root, "model.json"), "rb") as f:
        model_disk = json.load(f)

    model = json.loads(model.to_json())  # convert to str for comparison

    assert model == model_disk

    shutil.rmtree(root)


@pytest.mark.parametrize(
    "p",
    [
        model_h,
        model_p,
        model_hp,
        model_ht,
        model_pt,
        model_hpt,
    ],
)
def test_res_chkpt(p):
    model = Model.from_parameters(**p["model_params"])
    model.add_(**p["model_add_params"])
    # accommodate double-parameter case
    try:
        model.add_(**p["model_add_params2"])
    except KeyError:
        pass

    root = p["root_res"]
    executor = SparseSolver(System(model), root=root)
    w_grid = np.linspace(-3, -2, 10)

    results = executor.greens_function(p["k"], w_grid, eta=p["eta"]).squeeze()

    # load results from disk and confirm
    results_disk = []
    for w in w_grid:
        pathstr = executor._k_omega_eta_to_str(p["k"], w, p["eta"])
        ckpt_path = f"{pathstr}.pkl"
        path = executor._results_directory / Path(ckpt_path)
        results_disk.append(np.array(pickle.load(open(path, "rb"))))

    assert (results == results_disk).all()

    shutil.rmtree(root)



from pathlib import Path
import numpy as np
import time
import pickle

from petsc4py import PETSc

from ggce.executors.petsc4py.base import MassSolver
from ggce.logger import logger

BYTES_TO_GB = 1073741274


class MassSolverMUMPS(MassSolver):
    """A class to connect to PETSc powerful parallel sparse solver tools, to
    calculate G(k,w) in parallel, using a one-shot sparse solver MUMPS.
    This inherits the matrix construction strategies of the BaseExecutorPETSC
    base class.

    This is done because e.g. convergence and memory checks are often specific
    to the particular solver used -- and so is the KSP context (i.e. solver)
    setup."""

    def __init__(self, autoprime=True, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self._basis is None and autoprime:
            self._basis = self._system.get_basis(full_basis=True)

    def check_conv(self, factored_mat, rtol, elapsed):
        """This helper function checks MUMPS convergence using built-in MUMPS
        error codes. Factored_mat holds the MUMPS error codes and control
        values.

        Parameters
        ----------
        factored_mat : PETSc_Mat
            The factorized linear system matrix from the PETSc'
            pre-condictioning context PC, obtained afetr the
            solver has been called.

        Returns
        -------
        The MUMPS convergence criterion is conducted in place, nothing is
        returned.
        """

        # MUMPS main convergence index -- if 0, all good
        self.mumps_conv_ind = factored_mat.getMumpsInfog(1)

        # do the MUMPS check on the head node
        if self.mpi_rank == 0:
            if self.mumps_conv_ind == 0:
                logger.debug(
                    "According to MUMPS diagnostics, call to MUMPS was "
                    f"successful. The calculation took {elapsed:.2f} sec."
                )
            elif self.mumps_conv_ind < 0:
                logger.error(
                    "A MUMPS error occured with MUMPS error code "
                    f"{self.mumps_conv_ind} See the MUMPS User Guide, Sec. 8, "
                    "for error  diagnostics. The calculation took "
                    f"{elapsed:.2f} sec."
                )
            elif self.mumps_conv_ind > 0:
                logger.warning(
                    "A MUMPS warning occured with MUMPS warning code "
                    f"{self.mumps_conv_ind} See the MUMPS User Guide, Sec. 8, "
                    "for error diagnostics. The calculation took "
                    f"{elapsed:.2f} sec."
                )

    def check_mem_use(self, factored_mat):
        """This helper function checks MUMPS memory usage with built-in MUMPS
        access.

        Parameters
        ----------
        factored_mat : PETSc_Mat
            The factorized linear system matrix from the PETSc'
            pre-condictioning context PC, obtained afetr the
            solver has been called.

        Returns
        -------
        The memory resuts are given to the logger, nothing is returned.
        """

        # Each rank reports their memory usage (in millions of bytes)
        self.rank_mem_used = factored_mat.getMumpsInfo(26) * 1e6 / BYTES_TO_GB
        logger.debug(
            f"Current rank MUMPS memory usage is {self.rank_mem_used:.02f} GB"
        )

        # set up memory usage tracking, report to the logger on head node only
        # total memory across all processes
        self.total_mem_used = (
            factored_mat.getMumpsInfog(31) * 1e6 / BYTES_TO_GB
        )
        if self.mpi_rank == 0:
            logger.debug(
                f"Total MUMPS memory usage is {self.total_mem_used:.02f} GB"
            )

    def _pre_solve(self, k, w, eta):
        result = None
        path = None
        if self._results_directory is not None:
            ckpt_path = f"{self._k_omega_eta_to_str(k, w, eta)}.pkl"
            path = self._results_directory / Path(ckpt_path)
            if path.exists():
                result = np.array(pickle.load(open(path, "rb")))
        return result, path

    def _post_solve(self, G, k, w, path):
        if -G.imag / np.pi < 0.0:
            logger.error(f"A(k,w) < 0 at k, w = ({k:.02f}, {w:.02f}")
        if self._results_directory is not None:
            pickle.dump(G, open(path, "wb"), protocol=pickle.HIGHEST_PROTOCOL)

    def solve(self, k, w, eta, rtol=1.0e-10):
        """Solve the sparse-represented system using PETSc's KSP context.
        Note that this method only returns values on MPI rank = 0. All other
        ranks will return None.

        Parameters
        ----------
        k : float
            The momentum quantum number point of the calculation.
        w : float
            The frequency grid point of the calculation.
        eta : float
            The artificial broadening parameter of the calculation.
        rtol : float, optional
            PETSc's relative tolerance (the default is 1.0e-10).

        Returns
        -------
        np.ndarray, dict
            The value of G and meta information, which in this case, is only
            specifically the time elapsed to solve for this (k, w) point
            using the PETSc KSP context and MUMPS memory and exit codes.
        """

        # first check if you already calculated this
        result, path = self._pre_solve(k, w, eta)
        if result is not None:
            return [result, {}]

        if self._matr_dir is None:
            self._scaffold(k, w, eta)
        else:
            self._scaffold_from_disk(k, w, eta, matr_dir=self._matr_dir)

        t0 = time.time()

        # Now construct the desired solver instance
        ksp = PETSc.KSP().create()

        # "preonly" for e.g. mumps and other external solvers
        ksp.setType("preonly")

        # Define the linear system matrix and its preconditioner
        ksp.setOperators(self._mat_X, self._mat_X)

        # Set preconditioner options (see PETSc manual for details)
        pc = ksp.getPC()
        pc.setType("lu")
        pc.setFactorSolverType("mumps")

        # Set tolerance and options
        ksp.setTolerances(rtol=rtol)
        ksp.setFromOptions()

        dt = time.time() - t0
        logger.debug("KSP and PC contexts initialized", elapsed=dt)

        # Call the solve method
        t0 = time.time()
        ksp.solve(self._vector_b, self._vector_x)
        dt = time.time() - t0

        # assemble the solution vector
        self._vector_x.assemblyBegin()
        self._vector_x.assemblyEnd()

        # call manual residual check, as well as check MUMPS INFO
        # if the MUMPS solver call was successful
        # to get access to MUMPS error codes, get the factored matrix from PC
        factored_mat = pc.getFactorMatrix()
        self.check_conv_manual(pc, rtol=rtol)
        self.check_conv(factored_mat, rtol=rtol, elapsed=dt)

        # now check memory usage
        self.check_mem_use(factored_mat)

        self._mpi_comm_brigadier.barrier()

        # for memory management, destroy the KSP context manually
        ksp.destroy()

        # The last rank has the end of the solution vector, which contains G
        # G is the last entry aka "the last equation" of the matrix
        # use a gather operation, called by all ranks, to construct the full
        # vector (currently not used but will be later)
        G_vec = self._mpi_comm_brigadier.gather(
            self._vector_x.getArray(), root=0
        )

        # since we grabbed the Green's func value, destroy the data structs
        # not strictly necessary
        # self._vector_x.destroy()
        # self._vector_b.destroy()
        # self._mat_X.destroy()

        # Now select only the final value from the array
        if self.brigade_rank == 0:
            G_val = G_vec[self.brigade_size - 1][-1]
        else:
            G_val = None

        # and bcast to all processes in your brigade
        # self._mpi_comm_brigadier.Barrier()
        # self._mpi_comm.Abort()
        G_val = self._mpi_comm_brigadier.bcast(G_val, root=0)

        # only checkpoint if you are the brigade commander
        if self.brigade_rank == 0:
            self._post_solve(G_val, k, w, path)

        return np.array(G_val), {
            "time": [dt],
            "mumps_exit_code": [self.mumps_conv_ind],
            "mumps_mem_tot": [self.total_mem_used],
            "manual_tolerance_excess": [self.tol_excess],
        }



"""Utilities corresponding to physical quantities."""

import numpy as np


def g0_delta_omega(
    delta, omega, lattice_constant, broadening, hopping, sgn=1.0, e0=0.0
):
    """The free particle Green's function in position/frequency space. Note
    that the frequency is a complex variable, corresponding to
    :math:`\\omega + i \\eta`. This function is equation 21 in this
    `PRB <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.104.035106>`_.
    The equation is

    .. math::

        g_0(\\delta, \\omega) = -i \\frac{\\left[-\\omega_\\eta / 2t + i
        \\sqrt{1 - (\\omega_\\eta/2t)^2}\\right]^{|\\delta|}}{\\sqrt{4t^2 -
        \\omega_\\eta^2}}

    where :math:`\\omega_\\eta = \\omega + i \\eta`. The broadening is
    :math:`\\eta`, the hopping :math:`t`, and the lattice constant :math:`a`.

    .. warning::

        The parameters ``sgn`` and ``e0`` are not used in the GGCE formalism,
        but are explained in `E. N. Economou, Green’s Functions in Quantum
        Physics (Springer-Verlag, Berlin, 1983)`.

    Additionally, when :math:`t=0`, there are two cases. If
    :math:`\\delta = 0`, then the result is analytically 0. Otherwise, the
    result ends up being :math:`(\\omega + i\\eta)^{-1}`.

    Parameters
    ----------
    delta : float
        The site index variable
    omega : complex
        The complex frequency variable
    lattice_constant : float
        The lattice constant: distance between neighboring sites
    broadening : float
        The artificial broadening term
    hopping : float
        The hopping strength
    sgn : float, optional
        Optional sign parameter. It is recommended that this remain unchanged
        from the default. Default is 1.
    e0 : float, optional
        Optional E0 parameter. It is recommended that this remain unchanged
        from the default. Default is 1.

    Returns
    -------
    complex
        Value of :math:`g_0`.
    """

    a = lattice_constant

    if hopping == 0.0:
        if delta != 0:
            return 0.0
        return 1.0 / (omega + broadening * 1j)

    omega = omega + broadening * 1j
    B = 2.0 * np.abs(hopping)
    x = (omega - e0) / B
    prefactor = -sgn * 1j / np.sqrt(B**2 - (omega - e0) ** 2)
    t1 = (-x + sgn * 1j * np.sqrt(1.0 - x**2)) ** np.abs(delta * a)
    return t1 * prefactor


def G0_k_omega(k, omega, lattice_constant, eta, hopping):
    """The free particle Green's function on a 1D lattice in standard
    momentum-frequency space. The equation is

    .. math::

        G_0(k, \\omega) = \\frac{1}{\\omega + i\\eta - \\varepsilon_k}

    where :math:`\\varepsilon_k = -2t \\cos(ka)`.

    Parameters
    ----------
    k : float
        The momentum variable
    omega : complex
        The complex frequency variable.
    lattice_constant : float
        The lattice constant: distance between neighboring sites
    broadening : float
        The artificial broadening term
    hopping : float
        The hopping strength

    Returns
    -------
    complex
        Value of :math:`G_0`.
    """

    return 1.0 / (
        omega + 1j * eta + 2.0 * hopping * np.cos(k * lattice_constant)
    )



import pytest

import numpy as np

from ..model import Model as _Model


@pytest.fixture
def ZeroTemperatureModel():
    return _Model.from_parameters(
        hopping=1.0,
        lattice_constant=1.0,
        temperature=0.0,
        phonon_max_per_site=None,
        dimension=1,
    )


@pytest.fixture
def FiniteTemperatureModel():
    return _Model.from_parameters(
        hopping=1.0,
        lattice_constant=1.0,
        temperature=0.5,
        phonon_max_per_site=None,
        dimension=1,
    )


@pytest.fixture
def Random1dPhononArray():
    ptype1 = (np.random.random(size=(4,)) + 1.5).astype(int)
    ptype2 = (np.random.random(size=(4,)) + 2.5).astype(int)
    ptype3 = (np.random.random(size=(4,)) + 3.5).astype(int)
    return np.array([ptype1, ptype2, ptype3])


@pytest.fixture
def Random2dPhononArray():
    ptype1 = (np.random.random(size=(4, 5)) + 1.5).astype(int)
    ptype2 = (np.random.random(size=(4, 5)) + 2.5).astype(int)
    ptype3 = (np.random.random(size=(4, 5)) + 3.5).astype(int)
    return np.array([ptype1, ptype2, ptype3])


@pytest.fixture
def Random3dPhononArray():
    ptype1 = (np.random.random(size=(4, 5, 6)) + 1.5).astype(int)
    ptype2 = (np.random.random(size=(4, 5, 6)) + 2.5).astype(int)
    ptype3 = (np.random.random(size=(4, 5, 6)) + 3.5).astype(int)
    return np.array([ptype1, ptype2, ptype3])


# add this so slow tests are skipped by default
def pytest_addoption(parser):
    parser.addoption(
        "--runslow", action="store_true", default=False, help="run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")


def pytest_collection_modifyitems(config, items):
    if config.getoption("--runslow"):
        # --runslow given in cli: do not skip slow tests
        return
    skip_slow = pytest.mark.skip(reason="need --runslow option to run")
    for item in items:
        if "slow" in item.keywords:
            item.add_marker(skip_slow)



# before we import numpy, set the number of threads for numpy multithreading
# typically most gains are achieved by 4-6 threads
import numpy as np

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(ggce_head_dir)
    import ggce

from mpi4py import MPI

from ggce.model import Model  # noqa: E402
from ggce.executors.serial import SerialDenseExecutor
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS

# run PETSc benchmark at cloud size cloud_ext with up to bosons_max bosons
cloud_ext = 10
bosons_min = 3
bosons_max = 7
bosons_step = 1
bosons_array = np.arange(bosons_min, bosons_max + 1, 1)

# number of iterations to generate speed error bars
iters = 5

# set system parameters
Omega = 1.25
coupling = 2.5
hopping = 0.1
eta = 0.005

# pick specific point to do a calculation for speed benchmarking
w = -3.0
k = 0.5 * np.pi

# loop over the range of allowed N_bosons to execute one computation (k,w)
# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

parallel_times = np.zeros((len(bosons_array), iters))
parallel_greens = np.zeros((len(bosons_array), iters), dtype="complex")

for ii, n in enumerate(bosons_array):

    model = Model()
    model.set_parameters(hopping=hopping)
    model.add_coupling(
        "EdwardsFermionBoson",
        Omega=Omega,
        M=cloud_ext,
        N=n,
        dimensionless_coupling=coupling,
    )

    # repeat the calculation iters times for error bar generation
    for jj in range(iters):
        executor = ParallelSparseExecutorMUMPS(model, "info", mpi_comm=COMM)
        executor.prime()

        result = executor.solve(k, w, eta)

        if COMM.Get_rank() == 0:
            G, meta = result
            parallel_times[ii, jj] = meta["time"][-1]
            parallel_greens[ii, jj] = G

# now use the head node to do the serial calculation
if COMM.Get_rank() == 0:

    serial_times = np.zeros((len(bosons_array), iters))
    serial_greens = np.zeros((len(bosons_array), iters), dtype="complex")
    # do the same serially on the head node
    for ii, n in enumerate(bosons_array):

        model = Model()
        model.set_parameters(hopping=hopping)
        model.add_coupling(
            "EdwardsFermionBoson",
            Omega=Omega,
            M=cloud_ext,
            N=n,
            dimensionless_coupling=coupling,
        )

        # repeat the calculation iters times for error bar generation
        for jj in range(iters):
            executor = SerialDenseExecutor(model, "info", mpi_comm=COMM)
            executor.prime()

            G, meta = executor.solve(k, w, eta)

            serial_times[ii, jj] = meta["time"][-1]
            serial_greens[ii, jj] = G

## now we print to disk (and plot, if have matplotlib) the two timings to compare
if COMM.Get_rank() == 0:

    # first calculate averages and standard error
    time_average_serial = np.average(serial_times, axis=-1)
    greens_average_serial = np.average(serial_greens, axis=-1)
    std_error_serial = [
        np.std(serial_times[ii, :], ddof=1) / np.sqrt(len(serial_times[ii, :]))
        for ii in range(len(bosons_array))
    ]
    time_average_parallel = np.average(parallel_times, axis=-1)
    greens_average_parallel = np.average(parallel_greens, axis=-1)
    std_error_parallel = [
        np.std(parallel_times[ii, :], ddof=1)
        / np.sqrt(len(parallel_times[ii, :]))
        for ii in range(len(bosons_array))
    ]
    xx = np.array(
        [
            time_average_serial,
            std_error_serial,
            time_average_parallel,
            std_error_parallel,
            greens_average_serial,
            greens_average_parallel,
        ]
    ).T
    np.savetxt(
        os.path.join(
            script_dir,
            f"speed_benchmark_M_{str(cloud_ext)}_N_{bosons_min}_{bosons_max}.txt",
        ),
        xx,
        header=f"serial time (s)    serial time error (s)    parallel time (s)    parallel time error (s)"
        f"    serial G    parallel G",
        fmt=[
            "%.3f %+.0ej",
            "%.3f %+.0ej",
            "%.3f %+.0ej",
            "%.3f %+.0ej",
            "%.5e %+.5ej",
            "%.5e %+.5ej",
        ],
        delimiter="    ",
    )

    # uncomment if speed benchmark visualization desired
    plt.scatter(
        bosons_array,
        time_average_serial,
        label=f"serial, dense (cont'd frac.), M = {cloud_ext}",
    )
    plt.scatter(
        bosons_array,
        time_average_parallel,
        label=f"parallel, sparse (PETSc), M = {cloud_ext}",
    )
    plt.errorbar(
        bosons_array,
        time_average_serial,
        std_error_serial,
        0,
        "none",
        color="b",
    )
    plt.errorbar(
        bosons_array,
        time_average_parallel,
        std_error_parallel,
        0,
        "none",
        color="r",
    )
    plt.xlabel("N bosons", fontsize=16)
    plt.xticks(size=12)
    plt.ylabel("CPU time, seconds", fontsize=16)
    plt.yticks(size=12)
    plt.title(
        f"Speed comparison for M={cloud_ext}, N=[{bosons_min} to {bosons_max}]",
        fontsize=18,
    )
    plt.legend(fontsize=14)
    # plt.show()
    plt.savefig(
        os.path.join(
            script_dir,
            f"benchmark_petsc_M_{str(cloud_ext)}_N_{str(bosons_min)}_{str(bosons_max)}.png",
        ),
        format="png",
        bbox_inches="tight",
    )



import numpy as np
from scipy.optimize import curve_fit

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os
import shutil
import time

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = (
    f"/mnt/c/Users/bobst/Documents/University_of_British_Columbia/"
    f"Physics/Mona_Berciu/Generalized_Green_function_cluster_expan"
    f"sion/new_lorentzfit"
)
data_dir = (
    f"/mnt/c/Users/bobst/Documents/University_of_British_Columbia/"
    f"Physics/Mona_Berciu/Generalized_Green_function_cluster_expan"
    f"sion/fin_T_hol"
)
sys.path.append(ggce_head_dir)
import ggce

# load the mpi communicator
from mpi4py import MPI

# load the Model and Executor classes to store parameters and manage the calculation
from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS
from ggce.executors.serial import SerialDenseExecutor
from ggce.utils.utils import lorentzian, peak_location_and_weight_scipy

server = f"lisa"
Mval = 3
Nval = 9
Mval2 = 3
Nval2 = 5
lam, om, t = 1, 0.5, sys.argv[1]
eta = 0.05
date = sys.argv[2]
aux = f"polaron_dispersion"
year = f"22"
krange = [
    "0.000",
    "0.008",
    "0.017",
    "0.025",
    "0.033",
    "0.041",
    "0.050",
    "0.058",
    "0.066",
    "0.074",
    "0.083",
    "0.091",
    "0.099",
    "0.107",
    "0.116",
    "0.124",
    "0.132",
    "0.141",
    "0.149",
    "0.157",
]

results_dir = os.path.join(
    data_dir,
    f"lam_{lam}",
    f"om_{om}",
    f"t_{t}",
    aux,
    f"{Mval}{Nval}_{Mval2}{Nval2}_{server}_{date}_{year}",
    f"pol_dis_HfT_{Mval}{Nval}_{Mval2}{Nval2}",
)

energy_k = np.zeros(len(krange))
energy_k_err = np.zeros(len(krange))
eta_k = np.zeros(len(krange))

fig0, ax0 = plt.subplots()
for ii, k in enumerate(krange):
    result_file = os.path.join(results_dir, f"all_results_k_{k}.txt")
    wrange, Arange = np.loadtxt(result_file, unpack=True, skiprows=1)

    # restrict wrange to last 10 points near peak
    wrange = wrange[-10:]
    Arange = Arange[-10:]

    fit_params, error = peak_location_and_weight_scipy(wrange, Arange, eta)

    wrange_pred = np.linspace(wrange[0], wrange[-1], 100)
    Arange_pred = lorentzian(wrange_pred, *fit_params)

    if ii in [0, 10, 19]:
        ax0.plot(wrange_pred, Arange_pred)
        ax0.scatter(wrange, Arange)

    energy_k[ii] = fit_params[0]
    energy_k_err[ii] = error[0, 0]
    eta_k[ii] = fit_params[2]

# show together the plots of the lorentz fits
plt.show()

result_file = os.path.join(results_dir, f"polaron_dispersion.txt")
kgrid, energy_k_2point = np.loadtxt(result_file, unpack=True, skiprows=1)


def quadratic_band(k, m_star, offset):

    return k**2 / (2.0 * m_star) - offset


""" DOES ANYTHING CHANGE WHEN I FIT TO DOUBLE THE RANGE (SYMMETRIZE)?
NO, IT DOES NOT -- TINY EFFECTS IN THE 2-4 DECIMAL, NON-CRUCIAL """
# kgrid = np.array(list(kgrid)[::-1] + list(kgrid)[1:])
# energy_k = np.array( list(energy_k)[::-1] + list(energy_k)[1:] )
# energy_k_2point = np.array( list(energy_k_2point)[::-1] + list(energy_k_2point)[1:] )
# energy_k_err = np.array( list(energy_k_err)[::-1] + list(energy_k_err)[1:] )

best_fit_pars, covmat = curve_fit(
    quadratic_band, kgrid, energy_k, p0=[0.5, energy_k[0]]
)
best_fit_pars_2point, covmat_2point = curve_fit(
    quadratic_band, kgrid, energy_k_2point, p0=[0.5, energy_k_2point[0]]
)
dense_kgrid = np.linspace(0, kgrid[-1], 100)

""" IF YOU NEED TO SAVE THE SCIPY FITS (FOR EXAMPLE FOR PLOTTING PRETTIER E_p(K) BAND TRACKING) """
xx = np.array([kgrid, energy_k]).T
save_loc = os.path.join(results_dir, f"polaron_dispersion_scipy.txt")
np.savetxt(
    save_loc,
    xx,
    header=f"momentum        E_P(k) with m_star / m_0 = {best_fit_pars[0] / 0.5041061511616247} (m_0 = 0.5041061511616247)",
    delimiter="    ",
)
exit()
fig, ax = plt.subplots()
ax.plot(
    dense_kgrid,
    quadratic_band(dense_kgrid, *best_fit_pars),
    label=f"$k^2$ fit",
)

ax.plot(kgrid, energy_k_2point, label="two-point")
# plt.plot(krange_floats, all_locs, label = "scipy")
ax.errorbar(kgrid, energy_k, yerr=energy_k_err, label="scipy")
fig.legend()
plt.show()

print(
    f"scipy fit: m_star / m_0 = {best_fit_pars[0] / 0.5041061511616247} (m_0 = 0.5041061511616247)"
)
print(
    f"two-p fit: m_star / m_0 = {best_fit_pars_2point[0] / 0.5041061511616247} (m_0 = 0.5041061511616247)"
)
# # only collect and output results on the head node
# if COMM.Get_rank() == 0:
#
#     # plot and save the Lorentz fit for diagnostic purposes
#     for ii in range(len(kgrid)):
#         fig, ax = plt.subplots(1, 1)
#         loc = results[ii]["ground_state"]
#         scale = results[ii]["weight"]
#         ax.plot(results[ii]["w"], lorentzian(results[ii]["w"], loc, scale, eta), label = "Lorentz fit")
#         ax.scatter(results[ii]["w"], results[ii]["A"], label = "Calculation")
#         ax.set_ylabel(r"$A(k,w)$", fontsize=20)
#         ax.set_xlabel(r"$w$", fontsize=20)
#         ax.tick_params(which = 'both', labelsize = 16)
#         ax.legend()
#         plt.tight_layout()
#         plt.savefig(os.path.join(results_dir,f'lorentzfit_k_{kgrid[ii]:.3f}.png'),\
#                         format='png')
#         plt.close(fig)
#
#     # process the solver output to extract time to solve and spectrum vals
#     fig, ax = plt.subplots(1, 1)
#     ax.scatter(kgrid, energy_k, label = 'PETSc')
#     ax.set_ylabel(r"$E_P(k)$", fontsize=20)
#     ax.set_xlabel(r"$k/\pi$", fontsize=20)
#     ax.tick_params(which = "both", labelsize = 16)
#
#     def quadratic_band(k, m_star, offset):
#
#         return k**2 / (2.*m_star) - offset
#
#     best_fit_pars, covmat = curve_fit(quadratic_band, kgrid, energy_k, \
#                                                     p0 = [0.5, energy_k[0]])
#     dense_kgrid = np.linspace(0, kgrid[-1], 100)
#     ax.plot( dense_kgrid, quadratic_band(dense_kgrid, *best_fit_pars), \
#                                                         label = f"$k^2$ fit")
#
#     # also output the energy data to disk for posterity / postprocess
#     xx = np.array([kgrid, energy_k]).T
#     np.savetxt(os.path.join(results_dir,f"polaron_dispersion.txt"), xx,\
#                 header = f"momentum        E_P(k) with m_star / m_0 = {best_fit_pars[0] / 0.5041061511616247} (m_0 = 0.5041061511616247)",\
#                 delimiter = '    ')
#
#     plt.legend(bbox_to_anchor=(1,1), loc="upper left", fontsize=16)
#     plt.tight_layout()
#     # plt.show()
#     plt.savefig(os.path.join(results_dir,f'polaron_dispersion.png'), format='png', bbox_inches='tight')
#
#     ## also save all data just in case
#     for ii, kval in enumerate(kgrid):
#         xx = np.array([results[ii]['w'], results[ii]['A']]).T
#         np.savetxt(os.path.join(results_dir,f"all_results_k_{kval:.2f}.txt"), xx,\
#                     header = f"omega            spectral function        for momentum = {kval}",\
#                     delimiter = '    ')



from copy import deepcopy, copy
import numpy as np
import math

from monty.json import MSONable

from ggce.logger import logger


def model_coupling_map(coupling_type, t, Omega, lam):
    """Returns the value for g, the scalar that multiplies the coupling in the
    Hamiltonian. Converts the user-input lambda value to this g. Uses
    pre-defined values for the dimensionless coupling to get g for a variety
    of pre-defined default models.

    Parameters
    ----------
    coupling_type : str
        The desired coupling type. Can be Holstein, Peierls, BondPeierls, or
        EdwardsFermionBoson.
    t : float
        The hopping strength.
    Omega : float
        The (Einstein) phonon frequency. Absolute value is taken, since it will
        be negative sometimes (in the case of TFD).
    lam : float
        The dimensionless coupling.

    Returns
    -------
    float
        The value for the coupling (g).

    Raises
    ------
    RuntimeError
        If an unknown coupling type is provided.
    """

    if coupling_type == "Holstein":
        return math.sqrt(2.0 * t * np.abs(Omega) * lam)
    elif coupling_type == "EdwardsFermionBoson":
        return lam
    elif coupling_type == "Peierls":
        return math.sqrt(t * np.abs(Omega) * lam / 2.0)
    elif coupling_type == "BondPeierls":
        return math.sqrt(t * np.abs(Omega) * lam)
    else:
        raise RuntimeError


class SingleTerm(MSONable):
    """Contains the shift indexes, dagger status, coupling strength and boson
    type of a single term. Particularly, this single term corresponds to
    a single element of the sum in equation (10) in this
    `PRB <https://journals.aps.org/prb/abstract/10.1103/
    PhysRevB.104.035106>`_,

    .. math::

        g \\sum_i c_i^\\dagger c_{i + \\psi} b^d_{i + \\phi}
    """

    @property
    def psi(self):
        """The shift term on :math:`c`.

        Returns
        -------
        numpy.array
        """

        return self._psi.copy()

    @psi.setter
    def psi(self, x):
        if not isinstance(x, np.ndarray):
            logger.critical(f"psi {x} must be of type numpy.array")
        self._psi = x

    @property
    def phi(self):
        """The shift term on :math:`b^d`.

        Returns
        -------
        numpy.array
        """

        return self._phi.copy()

    @phi.setter
    def phi(self, x):
        if not isinstance(x, np.ndarray):
            logger.critical(f"phi {x} must be of type numpy.array")
        self._phi = x

    @property
    def dag(self):
        """The "dagger status" of the term. "+" for creation operator and "-"
        for annihilation operator. Must be either "+" or "-".

        Returns
        -------
        str
        """

        return self._dag

    @dag.setter
    def dag(self, x):
        if x not in ["+", "-"]:
            raise logger.critical(f"Invalid choice for dag: {x}")
        self._dag = x

    @property
    def coupling(self):
        """The coupling strength of the term. Directly multiplies the operators
        in the Hamiltonian. This is :math:`g`.

        Returns
        -------
        float
        """

        return self._coupling

    @coupling.setter
    def coupling(self, x):
        if not isinstance(x, (int, float)):
            raise logger.critical(f"Coupling {x} must be int or float")
        self._coupling = x

    @property
    def phonon_index(self):
        """The index of the phonon provided.

        Returns
        -------
        int
        """

        return self._phonon_index

    @phonon_index.setter
    def phonon_index(self, x):
        if not isinstance(x, int):
            raise logger.critical(f"Phonon index {x} must be of type int")
        if x < 0:
            raise logger.critical(f"Phonon index {x} must be >= 0")
        self._phonon_index = x

    @property
    def phonon_frequency(self):
        """The frequency of the phonon provided.

        .. note::

            The ``phonon_frequency`` can be negative in the TFD formalism.

        Returns
        -------
        float
            Description
        """

        return self._phonon_frequency

    @phonon_frequency.setter
    def phonon_frequency(self, x):
        if not isinstance(x, (int, float)):
            raise logger.critical(f"Phonon frequency {x} must be int or float")
        self._phonon_frequency = x

    def __init__(
        self,
        coupling_type,
        psi,
        phi,
        dag,
        coupling,
        phonon_index,
        phonon_frequency,
    ):
        """Initializes the SingleTerm class."""

        self._coupling_type = coupling_type
        self.psi = psi
        self.phi = phi
        self.dag = dag
        self.coupling = coupling
        self.phonon_index = phonon_index
        self.phonon_frequency = phonon_frequency

    def __str__(self):
        return (
            f"{self._coupling_type}: {self.coupling:.02f} x ({self.psi} "
            f"{self.phi} {self.dag}) | {self.phonon_index} "
            f"({self.phonon_frequency:.02f})"
        )

    def __repr__(self):
        return self.__str__()


class Hamiltonian(MSONable):
    """Container for all relevant parameters for defining the Hamiltonian.
    Current available are `Holstein`, `EdwardsFermionBoson`, `BondPeierls`,
    and `Peierls`."""

    # Old x is now psi
    # Old y is now phi
    # Old d is now dag
    # Old g is now sign; g must be provided elsewhere
    # Same as the PRB!

    @property
    def terms(self):
        return self._terms

    @property
    def phonon_frequencies(self):
        return self._phonon_frequencies

    @property
    def dimension(self):
        return self._dimension

    def get_dict_rep(self):
        """Provides a dictionary representation of the Hamiltonian. Packages
        terms by the phonon type.

        Returns
        -------
        dict
        """

        d = dict()
        for term in self._terms:
            try:
                d[term.phonon_index].append(term)
            except KeyError:
                d[term.phonon_index] = [term]
        return d

    def _get_SingleTerm_objects(
        self, coupling_type, coupling_strength, phonon_index, phonon_frequency
    ):
        """Gets the SingleTerm objects corresponding to some coupling_type,
        coupling strength and phonon_index."""

        # Holstein is easy. In all dimensions, it "looks" the same.
        if coupling_type == "Holstein":
            return [
                SingleTerm(
                    coupling_type,
                    np.array([0] * self._dimension),
                    np.array([0] * self._dimension),
                    "+",
                    -coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([0] * self._dimension),
                    np.array([0] * self._dimension),
                    "-",
                    -coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
            ]

        elif coupling_type == "EdwardsFermionBoson":
            return [
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([1]),
                    "+",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([-1]),
                    "+",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([0]),
                    "-",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([0]),
                    "-",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
            ]

        elif coupling_type == "BondPeierls":
            return [
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([0.5]),
                    "+",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([0.5]),
                    "-",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([-0.5]),
                    "+",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([-0.5]),
                    "-",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
            ]

        elif coupling_type == "Peierls":
            return [
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([0]),
                    "+",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([0]),
                    "-",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([1]),
                    "+",
                    -coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([1]),
                    np.array([1]),
                    "-",
                    -coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([-1]),
                    "+",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([-1]),
                    "-",
                    coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([0]),
                    "+",
                    -coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
                SingleTerm(
                    coupling_type,
                    np.array([-1]),
                    np.array([0]),
                    "-",
                    -coupling_strength,
                    phonon_index,
                    phonon_frequency,
                ),
            ]

        raise RuntimeError(f"Unknown coupling type {coupling_type}")

    def __init__(
        self, terms=[], phonon_frequencies=[], dimension=1, hopping=1.0
    ):
        """Sets the list of terms as an empty list"""

        self._terms = deepcopy(terms)
        self._phonon_frequencies = copy(phonon_frequencies)
        self._dimension = dimension
        self._hopping = hopping

    def _add_(
        self,
        coupling_type,
        phonon_index,
        phonon_frequency,
        coupling_strength=None,
        dimensionless_coupling_strength=None,
        coupling_multiplier=1.0,
    ):
        """Helper method for adding a term to the attribute. See `add_` for
        more details."""

        # XOR -- this should be taken care of outside this func
        c1 = coupling_strength is not None
        c2 = dimensionless_coupling_strength is not None
        if not c1 ^ c2:
            raise ValueError("Coupling strength errors")

        if c2:
            try:
                g = model_coupling_map(
                    coupling_type,
                    self._hopping,
                    phonon_frequency,
                    dimensionless_coupling_strength,
                )
            except RuntimeError:
                logger.error(f"Unknown coupling type {coupling_type}")
                return
        else:
            g = coupling_strength
        assert g is not None

        try:
            terms = self._get_SingleTerm_objects(
                coupling_type,
                g * coupling_multiplier,
                phonon_index,
                phonon_frequency,
            )
        except RuntimeError:
            logger.error(f"Unknown coupling type {coupling_type}")
            return

        if phonon_frequency not in self._phonon_frequencies:
            self._phonon_frequencies.append(phonon_frequency)

        self._terms.extend(terms)

    def add_(
        self,
        coupling_type,
        phonon_index,
        phonon_frequency,
        coupling_strength=None,
        dimensionless_coupling_strength=None,
        coupling_multiplier=1.0,
    ):
        """Adds a set of terms to the list of terms, corresponding to the
        provided coupling type, strength and index.

        Parameters
        ----------
        coupling_type : str
            The type of coupling to add to the Hamiltonian.
        phonon_index : int
            The index of the phonon to register.
        phonon_frequency : float
            The Einstein phonon frequency corresponding to the phonon creation
            and annihilation operators. Traditionally denoted :math:`\\Omega`.
        coupling_strength : float
            The strength of the coupling to add. This is g. Default is None.
            If None, requires `dimensionless_coupling_strength` to be set.
        dimensionless_coupling_strength : float
            The dimensionless coupling strength, which is a function of the
            hopping, and phonon frequency, as well as the type of model.
            Default is None. If None, requires `coupling_strength` to be set.
        coupling_multiplier : float, optional
            Extra term which will multiply the coupling. This could be a
            prefactor from e.g. the TFD formalism. Default is 1.0
        """

        args = {key: value for key, value in locals().items() if key != "self"}
        self._add_(
            coupling_type,
            phonon_index,
            phonon_frequency,
            coupling_strength,
            dimensionless_coupling_strength,
            coupling_multiplier,
        )
        logger.debug(f"Added {args}")

    def __str__(self):
        xx = "\n".join([str(xx) for xx in self._terms])
        return xx.replace(r"\n", "\n")

    def __repr__(self):
        return self.__str__()

    def visualize(self):
        print(self.__str__())


class Model(MSONable):
    """The Model object. Contains all information to be fed to the System
    class for solving for the Green's function.

    .. warning::

        Note it is highly recommended to use the :class:`from_parameters`
        classmethod instead, since the constructor is primarily designed to be
        used with MSONable and not all parameters are intended to be manually
        modified.
    """

    @property
    def temperature(self):
        """The temperature for a TFD simulation, if requested (the default is
        0.0).

        Returns
        -------
        float
        """
        return self._temperature

    @temperature.setter
    def temperature(self, x):
        if not isinstance(x, (float, int)):
            logger.error(f"Temperature {x} must be a number")
            return
        if x < 0.0:
            logger.error(f"Temperature {x} must be non-negative")
            return
        self._temperature = x

    @property
    def hopping(self):
        """The nearest-neighbor hopping term (the default is 1.0).

        Returns
        -------
        float
        """

        return self._hamiltonian._hopping

    @property
    def lattice_constant(self):
        """The lattice constant (the default is 1.0).

        Returns
        -------
        float
        """

        return self._lattice_constant

    @lattice_constant.setter
    def lattice_constant(self, x):
        if not isinstance(x, (float, int)):
            logger.error(f"Lattice constant {x} must be a number")
            return
        if x <= 0.0:
            logger.error(f"Lattice constant {x} must be positive")
            return
        if x != 1.0:
            logger.warning(f"Lattice constant {x}!=1, which is unusual")
        self._lattice_constant = x

    @property
    def phonon_absolute_extent(self):
        """The absolute extent value. Controls how much the clouds can overlap.
        Indicates the largest possible cloud size when multi-phonon mode
        calculations are run.

        Returns
        -------
        int
        """

        if self._phonon_absolute_extent is None:
            if len(self._phonon_extent) > 0:
                return np.max(self._phonon_extent).item()
            else:
                return None
        return self._phonon_absolute_extent

    @phonon_absolute_extent.setter
    def phonon_absolute_extent(self, x):
        L = len(self._phonon_extent)
        if L > 1:
            logger.error(f"{L} > 2 unique phonons, absolute extent not set")
            return
        self._phonon_absolute_extent = x

    @property
    def n_phonon_types(self):
        """A counter which keeps track of the number of phonon types that are
        contained in the model

        Returns
        -------
        int
        """

        return self._n_phonon_types

    @n_phonon_types.setter
    def n_phonon_types(self, x):
        assert isinstance(x, int)
        self._n_phonon_types = x

    @property
    def phonon_max_per_site(self):
        """Controls the hard-boson constraint. Essentially restrains the number
        of phonons that can be on each site. If None, there is no restriction.

        Returns
        -------
        int
        """

        return self._phonon_max_per_site

    @phonon_max_per_site.setter
    def phonon_max_per_site(self, x):
        if x is not None:
            assert isinstance(x, int)
        self._phonon_max_per_site = x

    @property
    def hamiltonian(self):
        """The Hamiltonian object for this model.

        Returns
        -------
        Hamiltonian
        """

        return self._hamiltonian

    @hamiltonian.setter
    def hamiltonian(self, x):
        assert isinstance(x, Hamiltonian)
        self._hamiltonian = x

    @property
    def phonon_extent(self):
        """The list of phonon extents. Commonly referred to as ``M``.

        Returns
        -------
        list
        """

        return self._phonon_extent

    @phonon_extent.setter
    def phonon_extent(self, x):
        assert x == []
        self._phonon_extent = []

    @property
    def phonon_number(self):
        """The list of max phonons. Commonly referred to as ``N``.

        Returns
        -------
        list
        """

        return self._phonon_number

    @phonon_number.setter
    def phonon_number(self, x):
        assert x == []
        self._phonon_number = []

    def visualize(self):
        """Visualize the model you've initialized. Note, some values are
        rounded."""

        print("Hamiltonian parameters:")
        print(f"  Hopping (t)          = {self.hopping:.02e}")
        print(f"  Lattice constant (a) = {self.lattice_constant:.02e}")
        print(f"  Temperature (T)      = {self.temperature:.02e}")
        print(f"  Max bosons per site  = {self.phonon_max_per_site}")
        print(f"  Absolute extent      = {self.phonon_absolute_extent}")
        if len(self._phonon_extent) == 0:
            print("Terms not initialized...")
            return
        print("Terms:")
        d1 = self.hamiltonian.get_dict_rep()
        for ii, (phonon_index, term_list) in enumerate(d1.items()):
            M = self._phonon_extent[ii]
            N = self._phonon_number[ii]
            print(f"  Phonon type = {phonon_index} (M = {M}; N = {N})")
            for term in term_list:
                print(f"    {term}")

    @classmethod
    def from_parameters(
        cls,
        hopping=1.0,
        lattice_constant=1.0,
        temperature=0.0,
        phonon_max_per_site=None,
        dimension=1,
    ):
        """Convenience method for initializing the Model object from intuitive
        parameters. It is recommended that this classmethod is used over the
        constructor.

        Parameters
        ----------
        hopping : float
            The hopping strength :math:`t`. Default is 1.0.
        lattice_constant : float, optional
            Default is 1.0.
        temperature : float, optional
            Note that setting this > 0 will engage the Thermofield Double
            method. Default is 0.0.
        phonon_max_per_site : None, optional
            Used for hard phonon constraints. Default is None, indicating no
            limitation on the number of phonons per site.
        dimension : int, optional
            The spatial dimension of the system to consider. Default is 1.
        """

        return cls(
            lattice_constant=lattice_constant,
            temperature=temperature,
            hamiltonian=Hamiltonian(dimension=dimension, hopping=hopping),
            phonon_max_per_site=phonon_max_per_site,
            phonon_extent=[],
            phonon_number=[],
            phonon_absolute_extent=None,
            n_phonon_types=0,
        )

    def __init__(
        self,
        lattice_constant=1.0,
        temperature=0.0,
        hamiltonian=Hamiltonian(),
        phonon_max_per_site=None,
        phonon_extent=[],
        phonon_number=[],
        phonon_absolute_extent=None,
        n_phonon_types=0,
    ):
        self._lattice_constant = lattice_constant
        self._temperature = temperature
        self._hamiltonian = deepcopy(hamiltonian)
        self._phonon_max_per_site = phonon_max_per_site
        self._phonon_extent = copy(phonon_extent)
        self._phonon_number = copy(phonon_number)
        self._phonon_absolute_extent = phonon_absolute_extent
        self._n_phonon_types = n_phonon_types

    def _get_TFD_coupling_prefactors(self, Omega):
        """Gets the TFD coupling prefactors.

        .. important::

            For reference, the TFD prefactors are defined in e.g.
            [JCP 145, 224101 (2016)](https://aip.scitation.org/doi/10.1063/
            1.4971211).

        Parameters
        ----------
        Omega : float
            The (Einstein) phonon frequency.

        Returns
        -------
        tuple of float
            The modifying prefactor to the real and fictitious couplings. If
            temperature is zero, then (1, 0) is returned.
        """

        if self.temperature > 0.0:
            beta = 1.0 / self.temperature
            theta_beta = np.arctanh(np.exp(-beta * Omega / 2.0))
            V_prefactor = np.cosh(theta_beta)
            V_tilde_prefactor = np.sinh(theta_beta)
            return V_prefactor, V_tilde_prefactor
        else:
            return 1.0, 0.0

    def add_(
        self,
        coupling_type,
        phonon_frequency,
        phonon_extent,
        phonon_number,
        phonon_extent_tfd=None,
        phonon_number_tfd=None,
        coupling_strength=None,
        dimensionless_coupling_strength=None,
        phonon_index_override=None,
    ):
        """Adds an electron-phonon contribution to the model. Note that the
        user can override the boson index manually to construct more
        complicated models, such as single phonon-mode Hamiltonians but with
        multiple contributions to the coupling_strength term.

        Parameters
        ----------
        coupling_type : str
            The type of coupling to add to the Hamiltonian.
        phonon_frequency : float
            The Einstein phonon frequency corresponding to the phonon creation
            and annihilation operators. Traditionally denoted :math:`\\Omega`.
        phonon_extent : int
            Positive number for the max phonon extent for this term.
        phonon_number : int
            Positive number for the max number of phonons for this term.
        phonon_extent_tfd : int, optional
            Positive number for the max phonon extent for the TFD version of
            the term. Default is None.
        phonon_number_tfd : int, optional
            Positive number for the max number of phonons for the TFD version
            of this term. Default is None.
        coupling_strength : float, optional
            The strength of the coupling to add. This is g. Default is None.
            If None, requires `dimensionless_coupling_strength` to be set.
        dimensionless_coupling_strength : float, optional
            The dimensionless coupling strength, which is a function of the
            hopping, and phonon frequency, as well as the type of model.
            Default is None. If None, requires `coupling_strength` to be set.
        phonon_index_override : int, optional
            Overrides the internal phonon counter. Use at your own risk.
            Note that for finite-temperature simulations, these indexes should
            always come in pairs. In other words, if the override 0, then
            the TFD part of the terms will be 1. Throws an error if this number
            is odd and temperature > 0. Default is None (no override).
        """

        # Make a few aliases for easier writing
        M = phonon_extent
        N = phonon_number
        M_tfd = phonon_extent_tfd
        N_tfd = phonon_number_tfd

        if self.temperature > 0.0 and (M_tfd is None or N_tfd is None):
            logger.error("Temperature > 0 but phonon extent or number not set")
            return

        if N < 1 or M < 1:
            logger.error(
                f"Provided phonon extent={M} and number={N} must be > 1"
            )
            return

        c1 = M_tfd is not None and M_tfd < 1
        c2 = N_tfd is not None and N_tfd < 1
        if c1 or c2:
            logger.error(
                f"Provided phonon M_tfd={M} and N_tfd={N} must be > 1"
            )
            return

        c1 = coupling_strength is None
        c2 = dimensionless_coupling_strength is None
        if not (c1 ^ c2):  # XOR
            logger.error(
                "One or the other of the coupling strength or dimensionless "
                "coupling strength should be set: XOR should be True"
            )
            return

        if phonon_index_override is not None and self.temperature > 0.0:
            if phonon_index_override % 2 != 0:
                logger.error("Phonon override is odd and temperature > 0")
                return

        if self.temperature == 0 and (M_tfd is not None or N_tfd is not None):
            logger.warning(
                "Temperature is set to zero but M_tfd or N_tfd values were "
                "provided and will be ignored."
            )

        # Get the TFD prefactors for the terms in the Hamiltonian. Note that
        # if temperature = 0, V_tilde_pf will be None.
        V_pf, V_tilde_pf = self._get_TFD_coupling_prefactors(phonon_frequency)

        # Add the standard term to the Hamiltonian
        if phonon_index_override is None:
            phonon_index = self._n_phonon_types
        else:
            phonon_index = phonon_index_override
        self._hamiltonian.add_(
            coupling_type,
            phonon_index,
            phonon_frequency,
            coupling_strength,
            dimensionless_coupling_strength,
            coupling_multiplier=V_pf,
        )
        self._phonon_extent.append(phonon_extent)
        self._phonon_number.append(phonon_number)
        self._n_phonon_types += 1
        phonon_index += 1

        # Add the finite-temperature term to the Hamiltonian
        if self.temperature > 0.0:
            self._hamiltonian.add_(
                coupling_type,
                phonon_index,
                -phonon_frequency,  # Negative phonon frequency!
                coupling_strength,
                dimensionless_coupling_strength,
                coupling_multiplier=V_tilde_pf,
            )

            assert phonon_extent_tfd is not None
            assert phonon_number_tfd is not None
            self._phonon_extent.append(phonon_extent_tfd)
            self._phonon_number.append(phonon_number_tfd)
            self._n_phonon_types += 1



#!/usr/bin/env python3

# Run with e.g.
# mpiexec -np `n_processes` python 001_ParallelExecutors.py
# to start in parallel

import numpy as np

from mpi4py import MPI

import sys
import os

script_dir = os.path.dirname(os.path.realpath(__file__))
head_ggce_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(head_ggce_dir)
    import ggce  # noqa: F401


# Import the parallel executor and load in the MPI communicator
from ggce.model import Model  # noqa: E402
from ggce.executors.parallel import ParallelDenseExecutor  # noqa: E402

COMM = MPI.COMM_WORLD

literature_data = np.loadtxt(os.path.join(script_dir, "000_example_A.txt"))

# Select the input parameters and load in the parallel model
model = Model()
model.set_parameters(hopping=0.1)
model.add_coupling(
    "EdwardsFermionBoson", Omega=1.25, M=3, N=9, dimensionless_coupling=2.5
)

executor = ParallelDenseExecutor(model, "info", mpi_comm=COMM)
executor.prime()

wgrid = np.linspace(-5.5, -2.5, 101)
spectrum = executor.spectrum(0.5 * np.pi, wgrid, eta=0.005)

# Results are returned on RANK 0 only
if COMM.Get_rank() == 0:
    xx = np.array([wgrid.squeeze(), spectrum.squeeze()]).T
    np.savetxt(os.path.join(script_dir, "tmp_parallel_results.txt"), xx)



############ Stress-testing PETSc ##########
"""
In this script we push PETSc to the limit. We will calculate entire spectral
slices, tracking both time to evaluate as well as saving the actual spectra.
We are looking to evalaute convergence wiht (M,N), as well as time to calculate
"""
########### ################

# before we import numpy, set the number of threads for numpy multithreading
# using the OMP_NUM_THREADS environment variable
import numpy as np

# uncomment if want to visualize PETSc speed benchmark
# import matplotlib as mpl
import matplotlib.pyplot as plt

import sys
import os
import time

script_dir = os.path.dirname(os.path.realpath(__file__))
ggce_head_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))

try:
    import ggce
except ModuleNotFoundError:
    sys.path.append(ggce_head_dir)
    import ggce

# load the mpi communicator
from mpi4py import MPI

# load the Model and Executor classes to store parameters and manage the calculation
from ggce.model import Model  # noqa: E402
from ggce.executors.petsc4py.parallel import ParallelSparseExecutorMUMPS

# run PETSc benchmark for cloud size up to cloud_max with up to bosons_max bosons
cloud_min = 3
cloud_max = 8
bosons_min = 9
bosons_max = 11

# fix parameter values same as ground truth from the https://journals.aps.org/prb/
# abstract/10.1103/PhysRevB.82.085116, figure 5 center for comparison
Omega = 1.25
coupling = 2.5
hopping = 0.1
eta = 0.005

# load the "grouth truth" data from https://journals.aps.org/prb/
# abstract/10.1103/PhysRevB.82.085116, figure 5 center for comparison
literature_data = np.loadtxt(
    os.path.join(ggce_head_dir, "examples", "000_example_A.txt")
)

# define w and k grid for the calculation
wgrid = np.linspace(-5.5, -2.5, 100)
k = 0.5 * np.pi

# loop over the range of allowed N_bosons to full spectrum slice (k,wgrid)
# start with defining an MPI communicator
COMM = MPI.COMM_WORLD

# start by creating a folder with the text and image output in run directory
# only head node make directories
results_dir = os.path.join(script_dir, "stress_test_results")
spectra_vis_path = os.path.join(
    script_dir, "stress_test_results", "spectra_visuals"
)
text_data_path = os.path.join(script_dir, "stress_test_results", "text_data")

if COMM.Get_rank() == 0:
    if not os.path.exists(results_dir):
        os.mkdir(results_dir)
    # now create nested directories for text and iamge respectively
    if not os.path.exists(spectra_vis_path):
        os.mkdir(spectra_vis_path)
    if not os.path.exists(text_data_path):
        os.mkdir(text_data_path)

# create an array to store time to finish the calculation
parallel_times = np.zeros((bosons_max, cloud_max))

# start the loop over all allowed M, N
for m in range(cloud_min, cloud_max + 1):
    for n in range(bosons_min, bosons_max + 1):

        # initialize the model with pre-determined parameters
        model = Model()
        model.set_parameters(hopping=hopping)
        model.add_coupling(
            "EdwardsFermionBoson",
            Omega=Omega,
            M=m,
            N=n,
            dimensionless_coupling=coupling,
        )

        # initialize the Executor to manage the calculation
        executor = ParallelSparseExecutorMUMPS(model, "info", mpi_comm=COMM)
        executor.prime()

        # execute the solve + time it
        t0 = time.time()
        results = [executor.solve(k, w, eta) for w in wgrid]
        dt = time.time() - t0
        # add the calculation time to the matrix
        parallel_times[n - 1, m - 1] = dt

        # only collect and output results on the head node
        if COMM.Get_rank() == 0:
            # process the solver output to extract time to solve and spectrum vals
            greens_func, times_as_dict = np.array(results).T
            spectrum = np.array([-s.imag / np.pi for s in greens_func])
            times = np.array(
                [time_dict["time"][0] for time_dict in times_as_dict]
            )

            # Plot to compare with "ground truth" (see https://journals.aps.org/prb/
            # abstract/10.1103/PhysRevB.82.085116, figure 5 center) via:
            fig, ax = plt.subplots(1, 1, figsize=(3, 2))
            ax.plot(
                wgrid,
                spectrum / spectrum.max(),
                "k",
                label=rf"PETSc ($M={m}, N={n}$)",
            )
            ax.plot(
                literature_data[:, 0],
                literature_data[:, 1] / literature_data[:, 1].max(),
                "r--",
                linewidth=0.5,
                label="Ground truth",
            )
            ax.set_ylabel("$A(\pi/2, \omega)$ [normalized]")
            ax.set_xlabel("$\omega$")
            plt.legend(bbox_to_anchor=(1, 1), loc="upper left")
            # plt.show()
            plt.savefig(
                os.path.join(
                    spectra_vis_path,
                    f"petsc_mumps_vs_groundtruth_M_{str(m)}_N_{str(n)}.png",
                ),
                format="png",
                bbox_inches="tight",
            )

            # also output the spectrum data to disk for posterity / postprocess
            xx = np.array([wgrid, spectrum, times]).T
            np.savetxt(
                os.path.join(
                    text_data_path,
                    f"parallel_results_M_{str(m)}_N_{str(n)}.txt",
                ),
                xx,
                header=f"omega      spectral func      time_to_compute",
                delimiter="    ",
            )


## now we print to disk the execution time for all combinations of M,N
if COMM.Get_rank() == 0:
    xx = parallel_times
    np.savetxt(
        os.path.join(
            results_dir,
            f"time_for_spectrum_M_{cloud_min}_{cloud_max}_N_{bosons_min}_{bosons_max}.txt",
        ),
        xx,
        header=f"parallel time (cloud size columns, boson number rows )",
        fmt="%.3f",
    )



"""The ``equations`` module constructs intermediate objects for dealing with
individual equations in the overall closure.

An :class:`.Equation` can be loosely thought of in the following
pseudo-mathematical equation:

.. math::

    f(n) \\sim \\sum_i f_i(n-1) + \\sum_j f_j(n+1) + b

where Auxiliary Green's functions, :math:`f`, are indexed by the total number
of phonons, :math:`n`. Note that this is of course a "bad" quantum number and
that much more information is required to represent the Auxiliary Green's
function, and that this information is actually contained, but for the sake
of explanation we suppress those parameters.

For couplings of the forms compatible with the GGCE software package, only
a single phonon is created or annihilated at a time. Thus, the equations in
the closer couple Auxiliary Green's Functions with :math:`n` phonons to those
with :math:`n \\pm 1` phonons. There is also a bias term, :math:`b`, which is
generally 0 except in the case of the :class:`.GreenEquation`, which is a
special instance of the :class:`.Equation` used only for the true Green's
function.

.. tip::

    During the operation of the GGCE code, it is unlikely you will need to
    actually manipulate the :class:`.Equation` and :class:`.GreenEquation`
    objects at all, so for the pure user, this module is likely not so
    necessary to understand.

.. note::

    See the :class:`.Equation` and :class:`.GreenEquation` for more details on
    these two important objects.
"""

from copy import deepcopy
import numpy as np

from monty.json import MSONable

from ggce.logger import logger
from ggce.utils import physics
from ggce.utils.utils import timeit
from ggce.engine import terms as terms_module


# TODO: MSONable is broken for this class and its children
class Equation(MSONable):
    """A single equation that indexes a single Auxiliary Green's Function
    (AGF). These functions can be loosely conceptualized in terms of the number
    of phonons in each term, and can be written as

    .. math::

        f(n) \\sim \\sum_i f_i(n-1) + \\sum_j f_j(n+1) + b

    where :math:`n` is an in-exhaustive index/bad quantum number corresponding
    to the total number of phonons on those sites. The :class:`Equation` class
    contains an ``index_term``, which in this case represents :math:`f(n)`, and
    a ``terms_list`` object, which corresponds to the right hand side.

    There is also a bias term, :math:`b`, which represents possible
    inhomogeneities. Only the Green's function itself contains
    :math:`b\\neq 0`. All of these terms are implicitly functions of the
    momentum, :math:`k`, the energy/frequency, :math:`\\omega`, and the
    artificial broadening :math:`\\eta`. For the base :class:`.Equation`, the
    ``bias`` is zero.

    Formally, this class is the representation of equation 22 in
    `PRB <https://journals.aps.org/prb/abstract/10.1103/PhysRevB.104.035106>`_.
    In one dimension, it is

    .. math::

        f_\\mathbf{n}(\\delta) = \\sum_{(g, \\psi, \\phi, \\xi)} g
        \\sum_{\\gamma \\in \\Gamma_L^\\xi} n^{(\\xi, \\gamma)}
        g_0(\\delta + \\gamma - \\phi + \\psi, \\tilde{\\omega})
        f_{\\mathbf{n}}^{(\\xi, \\gamma)}(\\phi - \\gamma)

    (see the manuscript for a detailed explanation of all of these terms).
    """

    @classmethod
    def from_config(cls, config, model):
        """Initializes the :class:`Equation` class from a numpy array and a
        :class:`ggce.models.Model`.

        Parameters
        ----------
        config : numpy.array
            The array representing the configuration of phonons.
        model : ggce.models.Model
            Container for the full set of parameters. Contains all Hamiltonian
            and parameter information about the terms, phonons, couplings, etc.

        Returns
        -------
        Equation
        """

        index_term = terms_module.IndexTerm(config.copy())
        return cls(index_term, model)

    @property
    def index_term(self):
        """The index term/config referencing the lhs of the equation.

        Returns
        -------
        ggce.engine.terms.IndexTerm
        """

        return self._index_term

    @property
    def model(self):
        """Container for the full set of parameters. Contains all Hamiltonian
        and parameter information about the terms, phonons, couplings, etc.

        Returns
        -------
        ggce.model.Model
            Description
        """

        return self._model

    @property
    def f_arg_terms(self):
        """A dictionary of lists of ints indexed by the config string
        representation containing lists of integers highlighting the required
        values of delta for that particular config string representation.

        Returns
        -------
        dict
        """

        return self._f_arg_terms

    def __init__(self, index_term, model, f_arg_terms=None, terms_list=None):
        logger.debug(f"Initializing {self.__class__.__name__} {index_term}")
        self._index_term = deepcopy(index_term)
        self._model = deepcopy(model)
        self._f_arg_terms = deepcopy(f_arg_terms)
        self._terms_list = deepcopy(terms_list)
        with timeit(logger.debug, "_initialize_terms"):
            self._initialize_terms()
        with timeit(logger.debug, "_populate_f_arg_terms"):
            self._populate_f_arg_terms()

    def bias(self, k, w, eta):
        """The default value for the bias is 0, except in the case of the
        Green's function."""

        return 0.0

    def visualize(self, full=True, coef=None):
        """Prints the representative strings of the terms for both the indexer
        and terms. This allows the user to at a high level, and
        without coefficients, visualize all the terms in the equation.
        If coef is not None, we assume we want to visualize the coefficient
        too, and coef is a 2-vector containing the k and w points.

        Parameters
        ----------
        full : bool, optional
            If True, prints out not only the configuration, its shape and the
            ``f_arg`` terms, but also the ``g_arg`` and ``shift`` components
            of the :class:`ggce.engine.terms.Term` object. Default is True.
        coef : None, optional
            Optional coefficients for :math:`k`, :math:`\\omega` and
            :math:`\\eta` to pass to the terms and output their coefficients.
            This is primarily used for debugging. Default is None (indicating
            no coefficient information will be printed).
        """

        id1 = self.index_term.id(full=full)
        if coef is not None:
            id1 += f" -> {self.index_term.coefficient(*coef)}"
        print(id1)
        if self._terms_list is not None:
            for term in self._terms_list:
                id1 = term.id(full=full)
                if coef is not None:
                    c = term.coefficient(*coef)
                    c_real = c.real.item()
                    c_imag = c.imag.item()
                    if c.imag < 0:
                        id1 += f" -> {c_real:.02e} - {-c_imag:.02e}i"
                    else:
                        id1 += f" -> {c_real:.02e} + {c_imag:.02e}i"
                print("\t", id1)

    def _populate_f_arg_terms(self):
        """Populates the required f_arg terms for the 'rhs'
        (in self._terms_list) that will be needed for the non-generalized
        equations later. Specifically, populates a dictionary all_delta_terms
        which maps the f_vec index string to a list of integers, which
        correspond to the needed delta values for that f_vec string. These
        dictionaries will be combined later during production of the
        non-generalized f-functions."""

        if self._f_arg_terms is not None:
            logger.error("f_args_terms is already initialized")
            return

        self._f_arg_terms = dict()
        for term in self._terms_list:
            n_mat_identifier = term._get_phonon_config_id()
            try:
                self._f_arg_terms[n_mat_identifier].append(term.f_arg)
            except KeyError:
                self._f_arg_terms[n_mat_identifier] = [term.f_arg]

    def _init_full(self, delta):
        """Increments every term in the term list's g-argument by the provided
        value. Also sets the f_arg to the proper value for the index term."""

        self.index_term._set_f_arg_(delta)
        for ii in range(len(self._terms_list)):
            self._terms_list[ii]._increment_g_arg_(delta)

    def _initialize_terms(self):
        """Systematically construct the generalized annihilation terms on the
        rhs of the equation."""

        if self._terms_list is not None:
            logger.error("terms_list is already initialized")
            return

        ae = self._model.phonon_absolute_extent
        n_spatial_dimensions = self._index_term.config.n_spatial_dimensions
        self._terms_list = []

        # Iterate over all possible types of the coupling operators
        for hterm in self._model.hamiltonian.terms:

            bt = hterm.phonon_index  # Phonon type
            arr = self.index_term.config.config[bt, ...]

            # We separate the two cases for creation and annihilation operators
            # on the boson operator 'b'
            if hterm.dag == "-":

                # Iterate over the site indexes for the term's alpha-index
                for loc, nval in np.ndenumerate(arr):

                    # Do not annihilate the vacuum, this term comes to 0
                    # anyway.
                    if nval == 0:
                        continue

                    t = terms_module.AnnihilationTerm(
                        self._index_term.config.config.copy(),
                        hamiltonian_term=hterm,
                        model=deepcopy(self._model),
                        constant_prefactor=hterm.coupling * nval,
                    )
                    t.step_(*loc)
                    t.check_if_green_and_simplify_()
                    t.config.validate()
                    if terms_module.config_legal(
                        t.config.config,
                        self._model.phonon_max_per_site,
                        self._model.phonon_extent,
                        allow_green=True,
                    ):
                        self._terms_list.append(t)

            elif hterm.dag == "+":

                # Don't want to create an equation corresponding to more than
                # the maximum number of allowed phonons.
                nphonons = self.index_term.config.total_phonons_per_type[bt]
                s = nphonons + 1
                if s > self._model.phonon_number[bt]:
                    continue

                if any([xx > ae for xx in self._index_term.config.shape[1:]]):
                    logger.critical(
                        f"Absolute extent {ae} cannot be smaller than any "
                        "config spatial dimension: "
                        f"{self._index_term.config.shape[1:]}"
                    )

                # Create a dummy array to make iterating over indexes easier
                dummy = np.empty(
                    [
                        2 * ae - self._index_term.config.shape[ii + 1]
                        for ii in range(n_spatial_dimensions)
                    ]
                )

                for _, (index, _) in enumerate(np.ndenumerate(dummy)):
                    loc = tuple(
                        [
                            index[jj]
                            - ae
                            + self._index_term.config.shape[jj + 1]
                            for jj in range(n_spatial_dimensions)
                        ]
                    )

                    t = terms_module.CreationTerm(
                        self._index_term.config.config.copy(),
                        hamiltonian_term=hterm,
                        model=deepcopy(self._model),
                        constant_prefactor=hterm.coupling,
                    )
                    t.step_(*loc)
                    t.check_if_green_and_simplify_()
                    t.config.validate()
                    if terms_module.config_legal(
                        t.config.config,
                        self._model.phonon_max_per_site,
                        self._model.phonon_extent,
                        allow_green=True,
                    ):
                        self._terms_list.append(t)


class GreenEquation(Equation):
    """Equation object corresponding to the Green's function. In one dimension,
    this corresponds directly with equation 13 in this
    `PRB <https://journals.aps.org/prb/abstract/10.1103/
    PhysRevB.104.035106>`_,

    .. math::

        f_0(0) - \\sum_{(g,\\psi,\\phi)} g e^{ik R_{\\psi - \\phi}} f_1(\\phi)
        = G_0(k, \\omega)

    The primary way in which this object differs from :class:`.Equation` is
    that it has a non-zero :class:`.GreenEquation.bias`.

    """

    def __init__(self, model):
        config = np.zeros(
            (
                model.n_phonon_types,
                *[1 for _ in range(model.hamiltonian.dimension)],
            )
        )
        index_term = terms_module.IndexTerm(config.copy())
        super().__init__(index_term=index_term, model=model)

    def bias(self, k, w, eta):
        """The bias term for the Green's function equation of motion.

        Parameters
        ----------
        k : float
            The momentum index.
        w : complex
            The complex frequency.
        eta : float
            The artificial broadening term.

        Returns
        -------
        complex
            This is :math:`G_0(k, \\omega; a, \\eta, t)`, where for the
            free particle Green's function on a lattice, :math:`G_0` is
            parameterized by the lattice constant, broadening and hopping.
        """

        return physics.G0_k_omega(
            k, w, self.model.lattice_constant, eta, self.model.hopping
        )

    def _initialize_terms(self):
        """Override for the Green's function other terms."""

        # Only instance where we actually use the base FDeltaTerm class
        self._terms_list = []

        # Iterate over all possible types of the coupling operators
        for hterm in self._model.hamiltonian.terms:

            # Only phonon creation terms contribute to the Green's function
            # EOM since annihilation terms hit the vacuum and yield zero.
            if hterm.dag == "+":
                n = self._index_term.config.config.copy()

                # This operation below is safe because the config for the
                # Green's function equation of motion will always have a shape
                # of (1, 1, ...).
                n[hterm.phonon_index, ...] = 1
                t = terms_module.EOMTerm(
                    n,
                    hamiltonian_term=hterm,
                    model=deepcopy(self._model),
                )
                self._terms_list.append(t)



#!/usr/bin/env python3

import numpy as np

import sys

sys.path.append("..")

from ggce import Model, System, DenseSolver  # noqa


if __name__ == "__main__":

    k = 0.5 * np.pi
    w = np.linspace(-2, 5, 50)

    # Check the true T=0 case
    model = Model.from_parameters(hopping=0.1, temperature=0.0)
    model.add_("Holstein", 1.25, 2, 2, dimensionless_coupling_strength=2.5)
    model.add_("Peierls", 2.25, 2, 2, dimensionless_coupling_strength=2.5)
    solver = DenseSolver(System(model))
    results = solver.spectrum(k, w, eta=0.005, pbar=False).squeeze()
    np.savetxt("T0.txt", np.array([w, -results.imag / np.pi]).T)

    # Check the true T=epsilon case
    model = Model.from_parameters(hopping=0.1, temperature=1e-6)
    model.add_(
        "Holstein",
        1.25,
        2,
        2,
        phonon_extent_tfd=2,
        phonon_number_tfd=3,
        dimensionless_coupling_strength=2.5,
    )
    model.add_(
        "Peierls",
        2.25,
        2,
        2,
        phonon_extent_tfd=2,
        phonon_number_tfd=3,
        dimensionless_coupling_strength=2.5,
    )
    solver = DenseSolver(System(model))
    results = solver.spectrum(k, w, eta=0.005, pbar=False).squeeze()
    np.savetxt("Tepsilon.txt", np.array([w, -results.imag / np.pi]).T)



from copy import deepcopy
import math
import pytest

import numpy as np

from ggce.model import model_coupling_map, SingleTerm, Hamiltonian


@pytest.mark.parametrize(
    "coupling_type,t,Omega,lam",
    [
        ("Holstein", 1.2345, 4.5678, 5.678),
        ("Holstein", 1.2345, -4.5678, 5.8),
        ("EdwardsFermionBoson", 1.2345, 4.5678, 5.678),
        ("Peierls", 1.2345, 4.5678, 5.678),
        ("BondPeierls", 1.2345, 4.5678, 5.678),
    ],
)
def test_model_coupling_map(coupling_type, t, Omega, lam):

    # This ground truth value is just copied from a version of the
    # model_coupling_map that is known to work
    def ground_truth(coupling_type, t, Omega, lam):
        if coupling_type == "Holstein":
            return math.sqrt(2.0 * t * np.abs(Omega) * lam)
        elif coupling_type == "EdwardsFermionBoson":
            return lam
        elif coupling_type == "Peierls":
            return math.sqrt(t * np.abs(Omega) * lam / 2.0)
        elif coupling_type == "BondPeierls":
            return math.sqrt(t * np.abs(Omega) * lam)
        else:
            raise RuntimeError

    assert model_coupling_map(coupling_type, t, Omega, lam) == ground_truth(
        coupling_type, t, Omega, lam
    )


class TestSingleTerm:
    @staticmethod
    @pytest.mark.parametrize(
        "coupling_type,psi,phi,dag,coupling,phonon_index,phonon_frequency",
        [
            ("Holstein", np.array([1]), np.array([-1]), "+", 1.23, 0, 4.56),
            ("Holstein", np.array([1]), np.array([-1]), "+", 1.23, 1, -4.56),
            ("Holstein", np.array([1]), np.array([-1]), "-", 1.23, 2, 4.56),
            ("Holstein", np.array([1]), np.array([-1]), "-", 1.23, 3, -4.56),
            ("Peierls", np.array([1]), np.array([-1]), "+", 1.23, 0, 4.56),
            (
                "BondPeierls",
                np.array([1]),
                np.array([-1]),
                "+",
                1.23,
                1,
                -4.56,
            ),
            (
                "EdwardsFermionBoson",
                np.array([1]),
                np.array([-1]),
                "-",
                1.23,
                2,
                4.56,
            ),
        ],
    )
    def test_initializer(
        coupling_type, psi, phi, dag, coupling, phonon_index, phonon_frequency
    ):
        SingleTerm(*[value for value in locals().values()])

    @staticmethod
    @pytest.mark.parametrize(
        "attribute,value",
        [
            ("psi", "hi"),
            ("phi", "hi"),
            ("dag", "not + or -"),
            ("dag", 1.0),
            ("dag", -2),
            ("coupling", "hi"),
            ("coupling", [1.0, -2.0, 3.0]),
            ("phonon_index", "hi"),
            ("phonon_index", -1),
            ("phonon_frequency", "hi"),
        ],
    )
    def test_setters_raises(attribute, value):
        st = SingleTerm(
            "Holstein", np.array([1]), np.array([-1]), "+", 1.23, 0, 4.56
        )
        with pytest.raises(SystemExit):
            setattr(st, attribute, value)


class TestHamiltonian:
    @staticmethod
    def test_initializer():
        Hamiltonian()

    @staticmethod
    @pytest.mark.parametrize(
        "coupling_type",
        ["Holstein", "Peierls", "EdwardsFermionBoson", "BondPeierls"],
    )
    @pytest.mark.parametrize("phonon_index", [0, 1, 2, -1])
    def test_add_(coupling_type, phonon_index):
        h = deepcopy(Hamiltonian())
        if phonon_index < 0:
            with pytest.raises(SystemExit):
                h.add_(coupling_type, phonon_index, 1.0, 1.0, None, 1.0)
        else:
            h.add_(coupling_type, phonon_index, 1.0, 1.0, None, 1.0)

            if coupling_type == "Holstein":
                assert len(h.terms) == 2
            elif coupling_type == "EdwardsFermionBoson":
                assert len(h.terms) == 4
            elif coupling_type == "BondPeierls":
                assert len(h.terms) == 4
            elif coupling_type == "Peierls":
                assert len(h.terms) == 8

    @staticmethod
    def test_invalid_coupling_add():
        h = deepcopy(Hamiltonian())
        h.add_("InvalidCoupling", 0, 2.0, 1.0, None, 1.0)
        assert len(h.terms) == 0

    @staticmethod
    @pytest.mark.parametrize(
        "coupling,dimensionless_coupling",
        [(1.0, 1.0), (None, None)],
    )
    def test_XOR_add_(coupling, dimensionless_coupling):
        h = deepcopy(Hamiltonian())
        with pytest.raises(ValueError):
            h.add_(
                coupling_type="Holstein",
                phonon_index=0,
                phonon_frequency=1.0,
                coupling_strength=coupling,
                dimensionless_coupling_strength=dimensionless_coupling,
                coupling_multiplier=1.0,
            )


class TestModel:
    @staticmethod
    def test_visualize(ZeroTemperatureModel):
        ZeroTemperatureModel.visualize()

    @staticmethod
    def test_visualize_after_add(ZeroTemperatureModel):
        ZeroTemperatureModel.add_("Holstein", 1.0, 2, 3)
        ZeroTemperatureModel.visualize()

    @staticmethod
    def test_finite_T_no_M_no_N_1(FiniteTemperatureModel):
        FiniteTemperatureModel.add_(
            coupling_type="Holstein",
            phonon_frequency=2.5,
            phonon_extent=3,
            phonon_number=2,
            phonon_extent_tfd=2,
            phonon_number_tfd=None,
            coupling_strength=1.0,
            dimensionless_coupling_strength=None,
            phonon_index_override=None,
        )
        assert len(FiniteTemperatureModel.phonon_number) == 0
        assert len(FiniteTemperatureModel.phonon_extent) == 0
        assert len(FiniteTemperatureModel.hamiltonian.terms) == 0

    @staticmethod
    def test_finite_T_no_M_no_N_2(FiniteTemperatureModel):
        FiniteTemperatureModel.add_(
            coupling_type="Holstein",
            phonon_frequency=2.5,
            phonon_extent=3,
            phonon_number=2,
            phonon_extent_tfd=None,
            phonon_number_tfd=3,
            coupling_strength=1.0,
            dimensionless_coupling_strength=None,
            phonon_index_override=None,
        )
        assert len(FiniteTemperatureModel.phonon_number) == 0
        assert len(FiniteTemperatureModel.phonon_extent) == 0
        assert len(FiniteTemperatureModel.hamiltonian.terms) == 0

    @staticmethod
    def test_finite_T_no_M_no_N_3(FiniteTemperatureModel):
        FiniteTemperatureModel.add_(
            coupling_type="Holstein",
            phonon_frequency=2.5,
            phonon_extent=3,
            phonon_number=2,
            phonon_extent_tfd=None,
            phonon_number_tfd=None,
            coupling_strength=1.0,
            dimensionless_coupling_strength=None,
            phonon_index_override=None,
        )
        assert len(FiniteTemperatureModel.phonon_number) == 0
        assert len(FiniteTemperatureModel.phonon_extent) == 0
        assert len(FiniteTemperatureModel.hamiltonian.terms) == 0



import pytest

from ggce.engine.system import config_space_gen


@pytest.mark.parametrize(
    "length,total_sum,expected",
    [
        (2, 2, {(2, 0), (1, 1), (0, 2)}),
        (2, 3, {(3, 0), (0, 3), (1, 2), (2, 1)}),
        (
            3,
            2,
            {(2, 0, 0), (0, 2, 0), (0, 0, 2), (1, 1, 0), (1, 0, 1), (0, 1, 1)},
        ),
    ],
)
def test_config_space_gen(length, total_sum, expected):
    assert set(config_space_gen(length, total_sum)) == expected



import pytest

import numpy as np

from ggce import Model, System

mpi4py_imported = False
try:
    from mpi4py import MPI

    mpi4py_imported = True
except ImportError:
    pass

petsc_imported = False
try:
    from ggce.executors.petsc4py.solvers import MassSolverMUMPS

    petsc_imported = True
except ImportError:
    pass


ATOL = 1.0e-6
TEPS = 1.0e-6

model_h = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.0
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=2,
        dimensionless_coupling_strength=0.6,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_p = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.0
    ),
    "model_add_params": dict(
        coupling_type="Peierls",
        phonon_frequency=2.35,
        phonon_extent=2,
        phonon_number=2,
        dimensionless_coupling_strength=1.3,
    ),
    "k": 1.46,
    "eta": 0.5,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_hp = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=0.0
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=2,
        dimensionless_coupling_strength=0.6,
    ),
    "model_add_params2": dict(
        coupling_type="Peierls",
        phonon_frequency=0.69,
        phonon_extent=2,
        phonon_number=2,
        dimensionless_coupling_strength=2.3,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_ht = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=TEPS
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=2,
        dimensionless_coupling_strength=0.6,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_pt = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=TEPS
    ),
    "model_add_params": dict(
        coupling_type="Peierls",
        phonon_frequency=2.35,
        phonon_extent=2,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=2,
        dimensionless_coupling_strength=1.3,
    ),
    "k": 1.46,
    "eta": 0.5,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}

model_hpt = {
    "model_params": dict(
        hopping=1.0, phonon_max_per_site=None, temperature=TEPS
    ),
    "model_add_params": dict(
        coupling_type="Holstein",
        phonon_frequency=1.23,
        phonon_extent=2,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=2,
        dimensionless_coupling_strength=0.6,
    ),
    "model_add_params2": dict(
        coupling_type="Peierls",
        phonon_frequency=0.69,
        phonon_extent=2,
        phonon_number=2,
        phonon_extent_tfd=2,
        phonon_number_tfd=2,
        dimensionless_coupling_strength=2.3,
    ),
    "k": 0.46,
    "eta": 0.005,
    "root_sys": "sys_chkpt",
    "root_res": "res_chkpt",
}


@pytest.mark.skipif(not petsc_imported, reason="PETSc not installed")
@pytest.mark.skipif(not mpi4py_imported, reason="mpi4py not installed")
@pytest.mark.mpi(min_size=2)
@pytest.mark.parametrize(
    "p",
    [
        [model_h, model_ht],
        [model_p, model_pt],
        [model_hp, model_hpt],
    ],
)
def test_zero_vs_tiny_T(p):

    COMM = MPI.COMM_WORLD

    k = p[0]["k"]
    w = np.linspace(-2, 0, 10)

    # Check the true T=0 case
    model = Model.from_parameters(**p[0]["model_params"])

    model.add_(**p[0]["model_add_params"])
    # accommodate double-model case
    try:
        model.add_(**p[0]["model_add_params2"])
    except KeyError:
        pass

    solver = MassSolverMUMPS(system=System(model), mpi_comm=COMM)
    results = solver.greens_function(k, w, eta=p[0]["eta"]).squeeze()

    # Check the true T=epsilon case
    model = Model.from_parameters(**p[1]["model_params"])

    model.add_(**p[1]["model_add_params"])
    # accommodate double-model case
    try:
        model.add_(**p[1]["model_add_params2"])
    except KeyError:
        pass

    solver = MassSolverMUMPS(system=System(model), mpi_comm=COMM)
    results_t = solver.greens_function(k, w, eta=p[1]["eta"]).squeeze()

    assert np.allclose(results, results_t, atol=ATOL)
